from itertools import count

from nmigen import *
from nap.util.python_misc import camel_to_snake

__all__ = ["Pipeline"]


class Pipeline:
    """A helper (syntactic sugar) to easier write pipelines of stream cores"""
    def __init__(self, m, prefix="", start_domain="sync"):
        self.m = m
        self.prefix = prefix
        self.next_domain = start_domain
        self.pipeline_members = {}

    def __setitem__(self, key, value):
        value = DomainRenamer(self.next_domain)(value)
        self.pipeline_members[key] = value
        self.m.submodules[key if not self.prefix else f'{self.prefix}_{key}'] = value

        # TODO: this concept breaks when a DomainRenamer is in the game;
        #       rethink how we handle this in that case
        if hasattr(value, "output_domain"):
            self.next_domain = value.output_domain

    def __iadd__(self, other):
        name = camel_to_snake(other.__class__.__name__)
        for i in count():
            n = name if i == 0 else f'{name}_{i}'
            if n not in self.pipeline_members:
                name = n
                break
        self.__setitem__(name, other)
        return self

    def __getitem__(self, item):
        return self.pipeline_members[item]

    @property
    def last(self):
        return list(self.pipeline_members.values())[-1]

    @property
    def output(self):
        return self.last.output


import unittest
from nap import PacketizedStream, write_packet_to_stream, SimPlatform
from nap.cores.video import PacketizedStream2ImageStream, read_frame_from_stream


class AdaptersTest(unittest.TestCase):
    def test_PacketizedStream2ImageStream(self):
        platform = SimPlatform()
        input_stream = PacketizedStream(32)
        dut = PacketizedStream2ImageStream(input_stream, width=10)

        def write_process():
            for frame in range(10):
                yield from write_packet_to_stream(input_stream, [0 for _ in range(100)])
        platform.add_process(write_process, "sync")

        def read_process():
            for frame in range(10):
                frame = yield from read_frame_from_stream(dut.output)
                self.assertEquals(len(frame), 10)
                self.assertTrue(all(len(l) == 10 for l in frame))
        platform.add_process(read_process, "sync")

        platform.add_sim_clock("sync", 100e6)
        platform.sim(dut)



from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum
from typing import TypeVar

from nmigen import *
from nmigen import tracer

from nap.util.python_misc import camel_to_snake

__all__ = ["Bundle", "Direction", "UPWARDS", "DOWNWARDS"]


class Bundle:
    def __init__(self, name=None, src_loc_at=1):
        super().__setattr__("name", name or tracer.get_var_name(depth=1 + src_loc_at, default="$" + camel_to_snake(self.__class__.__name__)))
        self._directions = OrderedDict()

    def __setattr__(self, key, value):
        if key == "name":
            super().__setattr__(key, value)
            self._update_name()

        if isinstance(value, Port):
            self._directions[key] = value.direction
            value = value.to_wrap
        elif isinstance(value, (Value, Bundle)):  # TODO: do we really want this behaviour also for values
            self._directions[key] = Direction.DOWNWARDS

        if hasattr(value, "name") and isinstance(value.name, str):
            if value.name.startswith("$"):  # with @UPWARDS and @DOWNWARDS we are breaking nmigens tracer
                value.name = key
            value.name = format("{}__{}".format(self.name, value.name))

        super().__setattr__(key, value)

    def __setitem__(self, key, value):
        assert isinstance(value, Port)
        self.__setattr__(key, value)

    def __getitem__(self, item):
        if item in self._directions:
            return getattr(self, item)
        else:
            raise KeyError(f"{item} not found in {repr(self)}")

    def _update_name(self):
        for attr_name in dir(self):
            attr = getattr(self, attr_name)
            if hasattr(attr, "name") and isinstance(attr.name, str):
                attr.name = format("{}__{}".format(self.name, attr.name.split("__")[-1]))

    def __repr__(self):
        return "<{}(Interface) name={}>".format(self.__class__.__name__, self.name)

    def connect_upstream(self, upstream, allow_partial=False, only=None, exclude=None):
        return self._connect(upstream, self, allow_partial, only, exclude)

    def connect_downstream(self, downstream, allow_partial=False, only=None, exclude=None):
        return self._connect(self, downstream, allow_partial, only, exclude)

    @property
    def signals(self):
        return [self[k] for k in self._directions.keys()]

    @staticmethod
    def _connect(upstream, downstream, allow_partial, only=None, exclude=None):
        if only is not None:
            assert exclude is None
            upstream_directions = {k: v for k, v in upstream._directions.items() if k in only}
            downstream_directions = {k: v for k, v in downstream._directions.items() if k in only}
        elif exclude is not None:
            assert only is None
            upstream_directions = {k: v for k, v in upstream._directions.items() if k not in exclude}
            downstream_directions = {k: v for k, v in downstream._directions.items() if k not in exclude}
        else:
            upstream_directions = upstream._directions
            downstream_directions = downstream._directions

        assert isinstance(upstream, Bundle) and isinstance(downstream, Bundle)
        if not allow_partial:
            assert upstream_directions == downstream_directions

        statements = []
        for k, direction in upstream_directions.items():
            if hasattr(upstream, k) and hasattr(downstream, k):
                assert upstream_directions[k] == downstream_directions[k]

                if direction == DOWNWARDS:
                    u, d = upstream[k], getattr(downstream, k)
                    if isinstance(d, Value):
                        statements += [d.eq(u)]
                    elif isinstance(d, Bundle):
                        statements += Bundle._connect(u, d, allow_partial)
                elif direction == UPWARDS:
                    u, d = upstream[k], downstream[k]
                    if isinstance(u, Value):
                        statements += [u.eq(d)]
                    elif isinstance(u, Bundle):
                        statements += Bundle._connect(d, u, allow_partial)
        return statements


class Direction(Enum):
    DOWNWARDS = 0
    UPWARDS = 1

    T = TypeVar('T')  # we lie about the return type to get nice IDE completions
    def __rmatmul__(self, other: T) -> T:
        return Port(self, other)


DOWNWARDS = Direction.DOWNWARDS
UPWARDS = Direction.UPWARDS


@dataclass
class Port:
    direction: Direction
    to_wrap: object



from nmigen import *
from nap.soc import SocPlatform, MemoryMap, Peripheral, Response

__all__ = ["SocMemory"]


class SocMemory(Elaboratable):
    """A memory that can be read / written to by the soc"""

    def __init__(self, *, width, depth, init=None, name=None, soc_read=True, soc_write=True):
        self.memory = Memory(width=width, depth=depth, init=init, name=name)
        self.soc_read = soc_read
        self.soc_write = soc_write
        self.depth = depth

    def elaborate(self, platform):
        m = Module()

        memory = m.submodules.memory = self.memory
        if isinstance(platform, SocPlatform):
            if self.soc_read:
                read_port = m.submodules.soc_read_port = memory.read_port(domain="sync", transparent=False)

                def handle_read(m, addr, data, read_done):
                    with m.If(addr > self.depth - 1):
                        read_done(Response.ERR)
                    with m.Else():
                        with m.FSM():
                            with m.State("ADDR"):
                                m.d.comb += read_port.addr.eq(addr)
                                m.d.comb += read_port.en.eq(1)
                                m.next = "DATA"
                            with m.State("DATA"):
                                m.d.comb += data.eq(read_port.data)
                                read_done(Response.OK)
                                m.next = "ADDR"
            else:
                def handle_read(m, addr, data, read_done):
                    read_done(Response.ERR)

            if self.soc_write:
                write_port = m.submodules.soc_write_port = memory.write_port(domain="sync", transparent=False)

                def handle_write(m, addr, data, write_done):
                    with m.If(addr > self.depth - 1):
                        write_done(Response.ERR)
                    with m.Else():
                        m.d.comb += write_port.addr.eq(addr)
                        m.d.comb += write_port.en.eq(1)
                        m.d.comb += write_port.data.eq(data)
                        write_done(Response.OK)
            else:
                def handle_write(m, addr, data, write_done):
                    write_done(Response.ERR)

            memorymap = MemoryMap()
            memorymap.allocate("drp", writable=True, bits=self.depth * memorymap.bus_word_width)

            m.submodules += Peripheral(
                handle_read,
                handle_write,
                memorymap
            )

        return m

    def read_port(self, *args, **kwargs):
        return self.memory.read_port(*args, **kwargs)

    def write_port(self, *args, **kwargs):
        return self.memory.write_port(*args, **kwargs)


import re
from inspect import getsource
from os.path import dirname, join
from textwrap import indent, dedent

from nmigen.build import Platform

from ..fatbitstream import FatbitstreamContext
from ..memorymap import MemoryMap
from ..tracing_elaborate import ElaboratableSames


def gen_hardware_proxy_python_code(mmap: MemoryMap, name="design", superclass="", top=True) -> str:
    name = name.lower()
    class_name = ("_" if not top else "") + name.capitalize()
    to_return = "class {}({}):\n".format(class_name, superclass)
    for row in mmap.direct_children:
        address = mmap.own_offset.translate(row.address)
        to_return += indent(
            "{} = (0x{:02x}, {}, {})\n".format(row.name, address.address, address.bit_offset, address.bit_len),
            "    "
        )
    for name, method in mmap.driver_methods.items():
        function_body = dedent(getsource(method.function))
        function_body_without_decorator = re.sub("^@.*$", "", function_body, flags=re.MULTILINE).strip()
        function_string = ("@property\n" if method.is_property else "") + function_body_without_decorator
        to_return += indent("\n" + function_string + "\n", "    ")

    for row in mmap.subranges:
        to_return += indent(gen_hardware_proxy_python_code(row.obj, row.name, superclass=superclass, top=False), "    ")
    return to_return


def generate_pydriver(top_memorymap, memory_accessor):
    pycode = "# pydriver hardware access file"
    pycode += "\n\n## HARDWARE PROXY STATIC: ###\n"
    pycode += open(join(dirname(__file__), "hardware_proxy.py")).read()
    pycode += "\n\n## HARDWARE PROXY DYNAMIC: ###\n"
    pycode += gen_hardware_proxy_python_code(top_memorymap, superclass="HardwareProxy")
    pycode += "\n\n## MEMORY ACCESSOR: ###\n"
    pycode += memory_accessor
    pycode += "\n\n## INTERACTIVE SHELL: ###\n"
    pycode += open(join(dirname(__file__), "interactive.py")).read()
    return pycode


def pydriver_hook(platform: Platform, top_fragment, sames: ElaboratableSames):
    if hasattr(platform, "pydriver_memory_accessor"):
        memorymap = top_fragment.memorymap
        pydriver = generate_pydriver(memorymap, platform.pydriver_memory_accessor)
        fc = FatbitstreamContext.get(platform)
        fc.self_extracting_blobs["pydriver.py"] = pydriver



# a helper for interfacing the ft601 usb3 fifo bridge in 245 Synchronous FIFO mode

from nmigen import *
from nap import Stream
from nap.cores import BufferedAsyncStreamFIFO

__all__ = ["FT601StreamSink", "FT601StreamSinkNoCDC"]


class FT601StreamSinkNoCDC(Elaboratable):
    def __init__(self, ft601_resource, input: Stream, save_to_begin_new_transaction=None):
        self.ft_601_resource = ft601_resource
        assert len(input.payload) == 32
        self.input = input
        self.safe_to_begin_new_transaction = save_to_begin_new_transaction

    def elaborate(self, platform):
        m = Module()

        ft = self.ft_601_resource

        m.d.comb += ft.be.o.eq(0b1111)  # everything we write is valid

        m.d.comb += ft.oe.eq(0)  # we are driving the data bits all the time

        if self.safe_to_begin_new_transaction is None:
            m.d.comb += ft.data.o.eq(self.input.payload)
            m.d.comb += self.input.ready.eq(ft.txe)
            m.d.comb += ft.write.eq(self.input.valid)
        else:
            in_transaction = Signal()
            m.d.sync += in_transaction.eq(ft.write)
            with m.If(in_transaction):
                m.d.comb += ft.write.eq(self.input.valid & ft.txe)
                m.d.comb += self.input.ready.eq(ft.txe)
            with m.Else():
                m.d.comb += self.input.ready.eq(ft.txe & self.safe_to_begin_new_transaction)
                m.d.comb += ft.write.eq(self.input.valid & self.safe_to_begin_new_transaction)
            m.d.comb += ft.data.o.eq(self.input.payload)

        return m


class FT601StreamSink(Elaboratable):
    def __init__(self, ft601_resource, input_stream, begin_transactions_at_level=2040, domain_name="ft601"):
        self.domain_name = domain_name
        self.ft601_resource = ft601_resource
        self.input_stream = input_stream
        self.begin_transactions_at_level = begin_transactions_at_level

    def elaborate(self, platform):
        m = Module()

        m.domains += ClockDomain(self.domain_name)
        m.d.comb += ClockSignal(self.domain_name).eq(self.ft601_resource.clk)

        cdc_fifo = m.submodules.cdc_fifo = BufferedAsyncStreamFIFO(
            self.input_stream, self.begin_transactions_at_level + 1, i_domain="sync", o_domain=self.domain_name
        )

        save_to_begin_new_transaction = Signal()
        m.d.comb += save_to_begin_new_transaction.eq(cdc_fifo.r_level >= self.begin_transactions_at_level)

        m.submodules.ft601 = DomainRenamer(self.domain_name)(FT601StreamSinkNoCDC(
            self.ft601_resource,
            cdc_fifo.output,
            save_to_begin_new_transaction,
        ))

        return m



# an experiment to test highspeed connectors for data integrity
from nmigen import *
from nmigen.build import Resource, Subsignal, DiffPairs, Attrs
from nap import *
from nap.vendor.xilinx_s7.clocking import _Pll, BufG
from nap.vendor.xilinx_s7.io import _OSerdes, _IDelayCtrl, _IDelay, _ISerdes


class Top(Elaboratable):
    def elaborate(self, platform: ZynqSocPlatform):
        platform.add_resources([
            Resource("loopback", 0,
                     # high speed serial lanes
                     Subsignal("tx", DiffPairs("1", "7", dir='o', conn=("pmod", "south")), Attrs(IOSTANDARD="LVDS_25")),
                     Subsignal("rx", DiffPairs("8", "2", dir='i', conn=("pmod", "south")), Attrs(IOSTANDARD="LVDS_25")),
                     )
        ])

        m = Module()

        # clock_setup: the serdes clockdomain has double the frequency of fclk1
        platform.ps7.fck_domain()
        pll = m.submodules.pll = _Pll(startup_wait=False, ref_jitter1=0.01, clkin1_period=8.0,
                                      clkfbout_mult=8, divclk_divide=1,
                                      clkout0_divide=16, clkout0_phase=0.0,
                                      clkout1_divide=4, clkout1_phase=0.0)
        platform.ps7.fck_domain(requested_frequency=200e6, domain_name="pll_clk_in")
        m.d.comb += pll.clk.in_[1].eq(ClockSignal("pll_clk_in"))
        m.d.comb += pll.clk.fbin.eq(pll.clk.fbout)
        bufg_serdes = m.submodules.bufg_serdes = BufG(pll.clk.out[0])
        m.domains += ClockDomain("serdes")
        m.d.comb += ClockSignal("serdes").eq(bufg_serdes.o)
        bufg_serdes_4x = m.submodules.bufg_serdes_4x = BufG(pll.clk.out[1])
        m.domains += ClockDomain("serdes_4x")
        m.d.comb += ClockSignal("serdes_4x").eq(bufg_serdes_4x.o)


        # make the design resettable via a axi register
        reset_serdes = ControlSignal()
        for domain in ["sync", "serdes", "serdes_4x"]:
            m.d.comb += ResetSignal(domain).eq(reset_serdes)

        # loopback
        loopback = platform.request("loopback")

        ## sender side
        to_oserdes = Signal(8)

        oserdes = m.submodules.oserdes = _OSerdes(
            data_width=8,
            tristate_width=1,
            data_rate_oq="ddr",
            serdes_mode="master",
            data_rate_tq="buf"
        )
        m.d.comb += oserdes.oce.eq(1)
        m.d.comb += oserdes.clk.eq(ClockSignal("serdes_4x"))
        m.d.comb += oserdes.clkdiv.eq(ClockSignal("serdes"))
        m.d.comb += oserdes.rst.eq(ResetSignal("serdes"))
        m.d.comb += Cat(oserdes.d[i] for i in reversed(range(1, 9))).eq(to_oserdes)  # reversed is needed!!1
        m.d.comb += loopback.tx.eq(oserdes.oq)

        ## reciver side
        bufg_idelay_refclk = m.submodules.bufg_idelay_refclk = BufG(pll.clk.out[1])
        m.domains += ClockDomain("idelay_refclk")
        m.d.comb += ClockSignal("idelay_refclk").eq(bufg_idelay_refclk.o)
        idelay_ctl = m.submodules.idelay_ctl = _IDelayCtrl()
        m.d.comb += StatusSignal(name="idelay_crl_rdy").eq(idelay_ctl.rdy)
        m.d.comb += idelay_ctl.refclk.eq(ClockSignal("idelay_refclk"))
        m.d.comb += idelay_ctl.rst.eq(ResetSignal("idelay_refclk"))
        idelay = m.submodules.idelay = _IDelay(
            delay_src="iDataIn",
            signal_pattern="data",
            cinvctrl_sel=False,
            high_performance_mode=True,
            refclk_frequency=200.0,
            pipe_sel=False,
            idelay_type="var_load",
            idelay_value=0
        )
        m.d.comb += idelay.c.eq(ClockSignal())  # this is really the clock to which the control inputs are syncronous!
        m.d.comb += idelay.ld.eq(ControlSignal(name="idelay_ld"))
        m.d.comb += idelay.ldpipeen.eq(0)
        m.d.comb += idelay.ce.eq(0)
        m.d.comb += idelay.inc.eq(0)
        m.d.comb += idelay.cntvalue.in_.eq(ControlSignal(5, name="idelay_cntvaluein"))
        m.d.comb += StatusSignal(5, name="idelay_cntvalueout").eq(idelay.cntvalue.out)
        m.d.comb += idelay.idatain.eq(loopback.rx)
        idelay_out = Signal()
        m.d.comb += idelay_out.eq(idelay.data.out)

        iserdes = m.submodules.iserdes = _ISerdes(
            data_width=8,
            data_rate="ddr",
            serdes_mode="master",
            interface_type="networking",
            num_ce=1,
            iobDelay="ifd",
        )
        m.d.comb += iserdes.ddly.eq(idelay_out)
        m.d.comb += iserdes.ce[1].eq(1)
        m.d.comb += iserdes.clk.eq(ClockSignal("serdes_4x"))
        m.d.comb += iserdes.clkb.eq(~ClockSignal("serdes_4x"))
        m.d.comb += iserdes.rst.eq(ResetSignal("serdes"))
        m.d.comb += iserdes.clkdiv.eq(ClockSignal("serdes"))
        from_iserdes = Signal(8)
        m.d.comb += from_iserdes.eq(Cat(iserdes.q[i] for i in range(1, 9)))

        ## check logic
        last_received = StatusSignal(8, name="last_received")

        current_state = StatusSignal(32, name="current_state")
        training_cycles = StatusSignal(32, name="training_cycles")
        slipped_bits = StatusSignal(32, name="slipped_bits")
        error_cnt = StatusSignal(32, name="error_cnt")
        success_cnt = StatusSignal(32, name="success_cnt")
        last_current_out = StatusSignal(24, name="last_current_out")
        test_pattern = ControlSignal(8, name="test_patern")

        with m.FSM(domain="serdes"):
            with m.State("TRAINING"):
                m.d.comb += current_state.eq(0)
                # start the link with correcting for the bitslip
                training_pattern = 0b00001111
                m.d.serdes += to_oserdes.eq(training_pattern)
                since_bitslip = Signal(2)
                m.d.serdes += training_cycles.eq(training_cycles + 1)
                with m.If(iserdes.bitslip == 1):
                    m.d.serdes += iserdes.bitslip.eq(0)
                    m.d.serdes += since_bitslip.eq(0)
                with m.Elif(since_bitslip < 3):
                    m.d.serdes += since_bitslip.eq(since_bitslip + 1)
                with m.Elif(from_iserdes != training_pattern):
                    m.d.serdes += iserdes.bitslip.eq(1)
                    m.d.serdes += slipped_bits.eq(slipped_bits + 1)
                with m.Else():
                    m.d.serdes += to_oserdes.eq(0x00)
                    m.next = "RUNNING"
            with m.State("RUNNING"):
                m.d.comb += current_state.eq(1)

                with m.If(test_pattern):
                    m.d.serdes += to_oserdes.eq(test_pattern)
                    with m.If(from_iserdes == test_pattern):
                        m.d.serdes += success_cnt.eq(success_cnt + 1)
                    with m.Else():
                        m.d.serdes += error_cnt.eq(error_cnt + 1)
                with m.Else():
                    m.d.serdes += to_oserdes.eq(to_oserdes + 1)
                    with m.If(from_iserdes == (last_received + 1)[0:8]):
                        m.d.serdes += success_cnt.eq(success_cnt + 1)
                    with m.Else():
                        m.d.serdes += error_cnt.eq(error_cnt + 1)

                m.d.serdes += last_received.eq(from_iserdes)
                m.d.comb += last_current_out.eq(Cat(last_received, from_iserdes, to_oserdes))

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(MicroR2Platform, ), possible_socs=(ZynqSocPlatform,))




from abc import ABC, abstractmethod
from textwrap import indent
from math import ceil, log2
from inspect import stack


class MemoryAccessor(ABC):
    base = 0

    @abstractmethod
    def read(self, addr):
        raise NotImplementedError()

    @abstractmethod
    def write(self, addr, value):
        raise NotImplementedError()


class BitwiseAccessibleInteger:
    def __init__(self, value=0):
        self.value = value

    def __getitem__(self, item):
        if isinstance(item, int):
            assert item >= 0
            return (self.value >> item) & 1
        elif isinstance(item, slice):
            assert 0 <= item.start <= item.stop
            assert (item.step == 1) or (item.step is None)
            bit_mask = (2**(item.stop - item.start) - 1)
            return (self.value >> item.start) & bit_mask

    def __setitem__(self, key, value):
        if isinstance(key, int):
            assert key >= 0
            assert value <= 1, "you can at maximum assign '1' to a 1 bit value"
            self.value = self.value & (((2**ceil(log2(self.value + 1))) - 1) ^ (1 << key))
            self.value = self.value | ((value & 1) << key)
        elif isinstance(key, slice):
            assert (key.step == 1) or (key.step is None)
            assert 0 <= key.start <= key.stop
            bit_mask = (2**(key.stop - key.start) - 1)
            assert value <= bit_mask, "you can at maximum assign '{}' to a {} bit value".format(bit_mask, (key.stop - key.start))
            self.value = self.value & (((2 ** ceil(log2(self.value + 1))) - 1) ^ (bit_mask << key.start))
            self.value = self.value | ((value & bit_mask) << key.start)

    def __int__(self):
        return self.value


class HardwareProxy:
    # TODO: test bit fiddeling
    def __init__(self, memory_accessor: MemoryAccessor):
        object.__setattr__(self, "_memory_accessor", memory_accessor)
        for k, v in self.__class__.__dict__.items():
            if isinstance(v, type) and issubclass(v, HardwareProxy):
                object.__setattr__(self, k[1:].lower(), v(memory_accessor))

    def __getattribute__(self, name):
        if name.startswith("__"):
            return object.__getattribute__(self, name)

        if name in {**self.__dict__, **self.__class__.__dict__}:
            obj = object.__getattribute__(self, name)

            if not isinstance(obj, tuple):
                return obj
            else:
                addr, bit_start, bit_len = obj

                to_return = BitwiseAccessibleInteger()
                read_bytes = ceil((bit_start + bit_len) / 32)
                for i in range(read_bytes):
                    val = BitwiseAccessibleInteger(self._memory_accessor.read(addr - self._memory_accessor.base + (i * 4)))
                    to_return[i*32:(i+1)*32] = val[bit_start:32+bit_start - (0 if i != read_bytes - 1 else bit_len % 32)]
                return int(to_return)
        raise AttributeError("{} has no attribute {}".format(self.__class__.__name__, name))

    def __setattr__(self, name, value):
        if name.startswith("__"):
            return object.__setattr__(self, name, value)

        if name in self.__class__.__dict__:
            obj = object.__getattribute__(self, name)

            if not isinstance(obj, tuple):
                return obj
            else:
                addr, bit_start, bit_len = obj
                assert value < 2**bit_len,  "you can at maximum assign '{}' to a {} bit value".format((2**bit_len - 1), bit_len)
                v = BitwiseAccessibleInteger(value)
                read_bytes = ceil((bit_start + bit_len) / 32)
                for i in range(read_bytes):
                    prev_value = BitwiseAccessibleInteger(self._memory_accessor.read(addr - self._memory_accessor.base + (i * 4)))
                    prev_value[bit_start:32+bit_start - (0 if i != read_bytes - 1 else bit_len % 32)] = v[i * 32:(i + 1) * 32]
                    self._memory_accessor.write(
                        addr - self._memory_accessor.base + (i * 4),
                        int(prev_value)
                    )
        else:
            raise AttributeError("{} has no attribute {}".format(self.__class__.__name__, name))

    def __repr__(self, allow_recursive=False):
        if stack()[1].filename == "<console>" or allow_recursive:
            to_return = ""
            children = [(name, getattr(self, name)) for name in dir(self) if not name.startswith("_")]
            real_children = [(name, child) for name, child in children if not isinstance(child, HardwareProxy)]
            proxy_children = [(name, child) for name, child in children if isinstance(child, HardwareProxy)]
            for name, child in real_children:
                if callable(child):
                    to_return += "{}: method()\n".format(name)
                else:
                    to_return += "{}: {}\n".format(name, child)
            for name, child in proxy_children:
                to_return += "{}: \n{}\n".format(name, indent(child.__repr__(allow_recursive=True), "    "))
            return to_return.strip()
        else:
            return "<HardwareProxy at 0x{:x}>".format(id(self))



from nmigen import *
from nap import BasicStream, StatusSignal
from .axi_endpoint import AxiResponse, AxiAddressStream, AxiBurstType
from .zynq_util import if_none_get_zynq_hp_port

__all__ = ["AxiReader"]


class AxiReader(Elaboratable):
    def __init__(
            self,
            address_source: BasicStream,
            axi=None, axi_data_width=64,
    ):
        self.address_source = address_source
        self.axi = axi

        self.output = BasicStream(axi_data_width, name="buffer_reader_output_stream")
        self.output.payload = Signal(axi_data_width)

        self.last_resp = StatusSignal(AxiResponse)
        self.error_count = StatusSignal(32)

    def elaborate(self, platform):
        m = Module()

        axi = if_none_get_zynq_hp_port(self.axi, m, platform)
        assert len(self.output.payload) == axi.data_bits
        assert len(self.address_source.payload) == axi.addr_bits
        for fifo_signal_name in ["read_address_fifo_level", "read_data_fifo_level"]:
            if hasattr(axi, fifo_signal_name):
                axi_fifo_signal = axi[fifo_signal_name]
                fifo_signal = StatusSignal(axi_fifo_signal.shape(), name=f"axi_{fifo_signal_name}")
                m.d.comb += fifo_signal.eq(axi_fifo_signal)
                setattr(self, f"axi_{fifo_signal_name}", fifo_signal)

        burster = m.submodules.burster = AxiReaderBurster(self.address_source, data_bytes=axi.data_bytes)
        m.d.comb += axi.read_address.connect_upstream(burster.output)
        m.d.comb += self.output.connect_upstream(axi.read_data, allow_partial=True)

        return m


class AxiReaderBurster(Elaboratable):
    """Creates Read Bursts. Outputs fewer (bursted) read addresses on its output that are equivalent to the input."""
    def __init__(
            self,
            address_source: BasicStream, data_bytes=8,
            max_burst_length=16, burst_creation_timeout=31,
    ):
        self.max_burst_length = max_burst_length
        self.burst_creation_timeout = burst_creation_timeout
        self.data_bytes = data_bytes
        self.input = address_source

        self.output = AxiAddressStream(addr_bits=len(self.input.payload), lite=False, id_bits=12, data_bytes=data_bytes)

    def elaborate(self, platform):
        m = Module()

        burst_ctr = Signal(range(self.max_burst_length + 1))
        timeout_ctr = Signal(range(self.burst_creation_timeout))
        burst_start = Signal.like(self.input.payload)
        last_address = Signal.like(self.input.payload)

        def write_address_burst():
            with m.If(burst_ctr > 0):
                m.d.comb += self.output.valid.eq(1)
                m.d.comb += self.output.payload.eq(burst_start)
                m.d.comb += self.output.burst_type.eq(AxiBurstType.INCR)
                m.d.comb += self.output.burst_len.eq(burst_ctr - 1)

        with m.If((self.input.payload == last_address + self.data_bytes) &
                  (burst_ctr < self.max_burst_length) &
                  self.input.valid):
            m.d.comb += self.input.ready.eq(1)
            m.d.sync += last_address.eq(self.input.payload)
            m.d.sync += burst_ctr.eq(burst_ctr + 1)
            m.d.sync += timeout_ctr.eq(0)
            with m.If(burst_ctr == 0):
                m.d.sync += burst_start.eq(self.input.payload)

        with m.Else():
            with m.If(self.input.valid):
                write_address_burst()
                with m.If(self.output.ready):
                    m.d.comb += self.input.ready.eq(1)
                    m.d.sync += last_address.eq(self.input.payload)
                    m.d.sync += burst_start.eq(self.input.payload)
                    m.d.sync += burst_ctr.eq(1)
                    m.d.sync += timeout_ctr.eq(0)
            with m.Elif((burst_ctr > 0)):
                with m.If(timeout_ctr < self.burst_creation_timeout - 1):
                    m.d.sync += timeout_ctr.eq(timeout_ctr + 1)
                with m.Else():
                    write_address_burst()
                    with m.If(self.output.ready):
                        m.d.comb += self.input.ready.eq(1)
                        m.d.sync += burst_ctr.eq(0)

        m.d.comb += self.output.id.eq(self.output.id.reset)
        m.d.comb += self.output.beat_size_bytes.eq(self.output.beat_size_bytes.reset)
        m.d.comb += self.output.protection_type.eq(self.output.protection_type.reset)

        return m


from nmigen import *

control_tokens = [0b1101010100, 0b0010101011, 0b0101010100, 0b1010101011]


class TmdsEncoder(Elaboratable):
    def __init__(self, data, control, data_enable):
        self.data = data
        self.control = control
        self.data_enable = data_enable

        self.out = Signal(10)

    def elaborate(self, platform):
        m = Module()

        # stage 1 - count number of 1s in data
        data = Signal(8)
        ones_in_data = Signal(range(9))
        m.d.sync += [
            ones_in_data.eq(sum(self.data[0:8])),
            data.eq(self.data)
        ]

        # stage 2 - add 9th bit
        q_m = Signal(9)
        q_m8_n = Signal()
        m.d.comb += q_m8_n.eq((ones_in_data > 4) | ((ones_in_data == 4) & ~data[0]))

        m.d.sync += q_m[0].eq(data[0])

        curval = data[0]
        for i in range(1, 8):
            curval = curval ^ data[i] ^ q_m8_n
            m.d.sync += q_m[i].eq(curval)
        m.d.sync += q_m[8].eq(~q_m8_n)

        # stage 3 - count number of 1s and 0s in q_m[:8]
        q_m_r = Signal(9)
        n0q_m = Signal(range(9))
        n1q_m = Signal(range(9))
        m.d.sync += [
            n0q_m.eq(sum(~q_m[0:8])),
            n1q_m.eq(sum(q_m[0:8])),
            q_m_r.eq(q_m)
        ]

        # stage 4 - final encoding
        disparity = Signal(signed(6))

        # match latency of control and data enable
        control = self.control
        data_enable = self.data_enable
        for p in range(3):
            new_c = Signal(2)
            new_de = Signal()
            m.d.sync += new_c.eq(control), new_de.eq(data_enable)
            control, data_enable = new_c, new_de

        with m.If(data_enable):
            with m.If((disparity == 0) | (n1q_m == n0q_m)):
                m.d.sync += [
                    self.out[9].eq(~q_m_r[8]),
                    self.out[8].eq(q_m_r[8])
                ]
                with m.If(q_m_r[8]):
                    m.d.sync += [
                        self.out[:8].eq(q_m_r[:8]),
                        disparity.eq(disparity + n1q_m - n0q_m)
                    ]
                with m.Else():
                    m.d.sync += [
                        self.out[:8].eq(~q_m_r[:8]),
                        disparity.eq(disparity + n0q_m - n1q_m)
                    ]
            with m.Else():
                with m.If((~disparity[5] & (n1q_m > n0q_m)) | (disparity[5] & (n0q_m > n1q_m))):
                    m.d.sync += [
                        self.out[9].eq(1),
                        self.out[8].eq(q_m_r[8]),
                        self.out[:8].eq(~q_m_r[:8]),
                        disparity.eq(disparity + Cat(0, q_m_r[8]) + n0q_m - n1q_m)
                    ]
                with m.Else():
                    m.d.sync += [
                        self.out[9].eq(0),
                        self.out[8].eq(q_m_r[8]),
                        self.out[:8].eq(q_m_r[:8]),
                        disparity.eq(disparity - Cat(0, ~q_m_r[8]) + n1q_m - n0q_m)
                    ]
        with m.Else():
            m.d.sync += [
                self.out.eq(Array(control_tokens)[control]),
                disparity.eq(0)
            ]

        return m


# A simple experiment that demonstrates basic CSR / SOC functionality
from nmigen import *
from nmigen.vendor.lattice_machxo_2_3l import LatticeMachXO2Platform
from nap import *
from nap.vendor.lattice_machxo2 import Osc


class Top(Elaboratable):
    def __init__(self):
        self.counter = StatusSignal(32)
        self.test_reg32 = ControlSignal(32)

    def elaborate(self, platform):
        m = Module()

        if isinstance(platform, ZynqSocPlatform):
            platform.ps7.fck_domain(requested_frequency=100e6)
            m.d.sync += self.counter.eq(self.counter + 1)
        elif isinstance(platform, LatticeMachXO2Platform):
            print("using lattice osc")
            m.submodules.osc = Osc()
            m.d.sync += self.counter.eq(self.counter + 1)
        else:
            m.d.comb += self.counter.eq(42)  # we dont have a clock source so we cant count

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(Usb3PluginPlatform, MicroR2Platform, ZyboPlatform, BetaPlatform, HdmiDigitizerPlatform, BetaRFWPlatform), possible_socs=(JTAGSocPlatform, ZynqSocPlatform))



# Test HDMI output using a given modeline by displaying a solid (adjustable) color

from nmigen import *
from nap import *


class Top(Elaboratable):
    def __init__(self):
        self.r = ControlSignal(8, reset=0xFA)
        self.g = ControlSignal(8, reset=0x87)
        self.b = ControlSignal(8, reset=0x56)

    def elaborate(self, platform: ZynqSocPlatform):
        if not isinstance(platform, ZyboPlatform):
            hdmi_plugin_connect(platform, "north")

        m = Module()

        hdmi_resource = platform.request("hdmi", "north")
        hdmi = m.submodules.hdmi = HdmiTx(hdmi_resource, generate_modeline(1920, 1080, 30))

        clocking_debug = m.submodules.clocking_debug = ClockingDebug("pix", "pix_5x")
        
        m.d.comb += hdmi.rgb.r.eq(self.r)
        m.d.comb += hdmi.rgb.g.eq(self.g)
        m.d.comb += hdmi.rgb.b.eq(self.b)

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(MicroR2Platform, BetaPlatform, ZyboPlatform), possible_socs=(ZynqSocPlatform, ))



from nmigen import *
from nap import RGB24


class BertlPatternGenerator(Elaboratable):
    def __init__(self, width, height):
        self.x = Signal.like(width)
        self.y = Signal.like(height)
        self.out = RGB24()

    def elaborate(self, platform):
        m = Module()

        m.d.comb += self.out.r.eq(self.x[0:8])
        m.d.comb += self.out.g.eq(self.y[0:8])
        m.d.comb += self.out.b.eq(Cat(Signal(3), self.y[8:10], self.x[8:11]))

        return m


class DimmingPatternGenerator(Elaboratable):
    def __init__(self, width, height):
        self.x = Signal(range(width))
        self.y = Signal(range(height))
        self.out = RGB24()

    def elaborate(self, platform):
        m = Module()

        frame_counter = Signal(range(256 * 3 + 1))
        with m.If((self.x == 0) & (self.y == 0) & (frame_counter < 256 * 3)):
            m.d.sync += frame_counter.eq(frame_counter + 1)
        with m.Elif((self.x == 0) & (self.y == 0)):
            m.d.sync += frame_counter.eq(0)

        with m.If(self.x < 256 * 1):
            m.d.comb += self.out.r.eq(self.x)
        with m.Elif(self.x < 256 * 2):
            m.d.comb += self.out.g.eq(self.x)
        with m.Elif(self.x < 256 * 3):
            m.d.comb += self.out.b.eq(self.x)

        return m


from nmigen import *
from nap import BasicStream, PacketizedStream, Stream, DOWNWARDS, StatusSignal
from . import BufferedSyncStreamFIFO, StreamTee, StreamCombiner

__all__ = ["LastWrapper", "GenericMetadataWrapper"]


class LastWrapper(Elaboratable):
    """wraps a core that transforms one BasicStream into another BasicStream (with variable latency) and applies the last signal correctly to its output"""

    # TODO: if we have a last_fifo_depth of 2 we cant do formal anymore because of this yosys bug:
    #       https://github.com/YosysHQ/yosys/issues/2577
    def __init__(self, input: PacketizedStream, core_producer, last_fifo_depth=2, last_rle_bits=10):
        self.last_fifo_depth = last_fifo_depth
        self.last_rle_bits = last_rle_bits
        assert hasattr(input, "last")
        self.input = input

        self.core_input = BasicStream(input.payload.shape())
        self.core = core_producer(self.core_input)
        self.core_output = self.core.output

        self.error = StatusSignal(32)

        self.output = PacketizedStream(len(self.core_output.payload))

    def elaborate(self, platform):
        m = Module()
        m.submodules.core = self.core

        last_fifo_input = BasicStream(self.last_rle_bits)
        last_fifo = m.submodules.last_fifo = BufferedSyncStreamFIFO(last_fifo_input, self.last_fifo_depth)

        overflow_word = (2 ** self.last_rle_bits - 1)

        rle_input_counter = StatusSignal(self.last_rle_bits)
        with m.If(self.input.valid & last_fifo_input.ready):
            m.d.comb += self.core_input.valid.eq(1)
            m.d.comb += self.core_input.payload.eq(self.input.payload)

            with m.If(self.core_input.ready):
                m.d.comb += self.input.ready.eq(1)

                with m.If(self.input.last | (rle_input_counter == overflow_word - 1)):
                    with m.If(~self.input.last & (rle_input_counter == overflow_word - 1)):
                        m.d.comb += last_fifo_input.payload.eq(overflow_word)
                    with m.Else():
                        m.d.comb += last_fifo_input.payload.eq(rle_input_counter)
                    m.d.comb += last_fifo_input.valid.eq(1)
                    m.d.sync += rle_input_counter.eq(0)
                with m.Else():
                    m.d.sync += rle_input_counter.eq(rle_input_counter + 1)

        rle_output_counter = StatusSignal(self.last_rle_bits)
        with m.If(self.core_output.valid):
            m.d.comb += self.output.valid.eq(1)
            m.d.comb += self.output.payload.eq(self.core_output.payload)

            overflow = (last_fifo.output.payload == overflow_word) & (rle_output_counter == (overflow_word - 1))
            with m.If(((rle_output_counter == last_fifo.output.payload) | overflow) & last_fifo.output.valid):
                with m.If(~overflow):
                    m.d.comb += self.output.last.eq(1)
                with m.If(self.output.ready):
                    m.d.sync += rle_output_counter.eq(0)
                    m.d.comb += last_fifo.output.ready.eq(1)
                    m.d.comb += self.core_output.ready.eq(1)
            with m.Elif((rle_output_counter > last_fifo.output.payload) & last_fifo.output.valid):
                with m.If(self.output.ready):
                    m.d.comb += self.core_output.ready.eq(1)
                    m.d.sync += self.error.eq(self.error + 1)
            with m.Else():
                with m.If(self.output.ready):
                    m.d.comb += self.core_output.ready.eq(1)
                    m.d.sync += rle_output_counter.eq(rle_output_counter + 1)

        return m


class GenericMetadataWrapper(Elaboratable):
    def __init__(self, input: BasicStream, core_producer, fifo_depth=128):
        self.fifo_depth = fifo_depth
        self.input = input

        self.core_input = BasicStream(input.payload.shape())
        self.core = core_producer(self.core_input)
        self.core_output = self.core.output

        self.output = self.input.clone()
        self.output.payload = Signal(len(self.core_output.payload)) @ DOWNWARDS

    def elaborate(self, platform):
        m = Module()
        m.submodules.core = self.core
        tee = m.submodules.tee = StreamTee(self.input)

        m.d.comb += self.core_input.connect_upstream(tee.get_output(), allow_partial=True)

        metadata_fifo_input = Stream()
        for name, s in self.input.out_of_band_signals.items():
            setattr(metadata_fifo_input, name, Signal.like(s) @ DOWNWARDS)
        m.d.comb += metadata_fifo_input.connect_upstream(tee.get_output(), allow_partial=True)
        metadata_fifo = m.submodules.metadata_fifo = BufferedSyncStreamFIFO(metadata_fifo_input, self.fifo_depth)

        combiner = m.submodules.combiner = StreamCombiner(self.core_output, metadata_fifo.output)
        m.d.comb += self.output.connect_upstream(combiner.output)

        return m



from typing import get_type_hints

from nmigen import *
from nmigen import tracer
from nmigen.hdl.ast import ValueCastable

from nap.util.python_misc import camel_to_snake

__all__ = ["packed_struct"]


class PackedStructBaseClass(ValueCastable):
    def __init__(self, backing_signal=None, name=None, src_loc_at=1, **kwargs):
        self.name = name or tracer.get_var_name(depth=2 + src_loc_at, default=camel_to_snake(self.__class__.__name__))

        if backing_signal is not None:
            assert isinstance(backing_signal, Value)
            assert len(kwargs) == 0
            assert len(backing_signal) == self._PACKED_LEN
            self._backing_signal = backing_signal
        else:
            # we do not create a single backing signal but instead cat multiple signals together for better
            # introspection in the generated code and vcd files
            def get_signal(name, type):
                if name in kwargs:
                    signal = kwargs[name]
                    if hasattr(type, "_PACKED_FIELDS"):
                        assert isinstance(signal, type)
                    assert needed_bits(type) == len(signal)
                    signal.name = f'{self.name}__{name}'
                    return signal
                elif hasattr(type, "_PACKED_FIELDS"):
                    return type(name=f'{self.name}__{name}')
                else:
                    return Signal(type, name=f'{self.name}__{name}')

            self._backing_signals = {
                name: get_signal(name, type)
                for name, type in self._PACKED_FIELDS.items()
            }
            self._backing_signal = Cat(self._backing_signals.values())

    def __getattribute__(self, item):
        if item in object.__getattribute__(self, "_PACKED_FIELDS"):
            if hasattr(self, "backing_signals"):
                # we have the individual components of the signal so we do not need to slice them out
                return self.backing_signals[item]
            else:
                # fallback path for when we are backed by a single signal
                start, stop = self._PACKED_SLICES[item]
                type = self._PACKED_FIELDS[item]
                if hasattr(type, "_PACKED_FIELDS"):
                    return type(self._backing_signal[start:stop])
                else:
                    return self._backing_signal[start:stop]
        else:
            return object.__getattribute__(self, item)

    def __len__(self):
        return self._PACKED_LEN

    @ValueCastable.lowermethod
    def as_value(self):
        return self._backing_signal

    def eq(self, other):
        assert isinstance(other, self.__class__)
        return self.as_value().eq(other.as_value())


def needed_bits(obj):
    if hasattr(obj, "_PACKED_LEN"):
        return obj._PACKED_LEN
    else:
        return Shape.cast(obj).width


def packed_struct(cls):
    """A decorator that turns a class into a packed struct (similiar to dataclass)"""
    cls._PACKED_FIELDS = get_type_hints(cls)  # this only works in python 3.6+ because dict ordering is not consistent before
    cls._PACKED_SLICES = {}
    last_index = 0
    for name, field_type in cls._PACKED_FIELDS.items():
        bits = needed_bits(field_type)
        cls._PACKED_SLICES[name] = (last_index, last_index + bits)
        last_index += bits
    cls._PACKED_LEN = last_index

    return type(cls.__name__, (PackedStructBaseClass,), dict(cls.__dict__))



import random
import unittest

from nap.cores.stream.buffer import StreamBuffer
from nap.stream.formal_util import verify_stream_output_contract
from nap.stream.sim_util import write_to_stream, read_from_stream
from nap.stream import BasicStream
from nap.util.sim import SimPlatform


class TestBuffer(unittest.TestCase):
    def test_basic(self):
        platform = SimPlatform()
        input_stream = BasicStream(32)
        dut = StreamBuffer(input_stream)

        random.seed(0)
        test_data = [random.randrange(0, 2**32) for _ in range(100)]

        def write_process():
            for d in test_data:
                yield from write_to_stream(input_stream, payload=d)
        platform.add_process(write_process, "sync")

        def read_process():
            for expected in test_data:
                read = yield from read_from_stream(dut.output)
                self.assertEquals(read, expected)
        platform.add_process(read_process, "sync")

        platform.add_sim_clock("sync", 100e6)
        platform.sim(dut)

    def test_output_properties(self):
        input = BasicStream(32)
        # important: we prove here that ANY input produces a contract obeying output
        verify_stream_output_contract(StreamBuffer(input))



import unittest
from nmigen import *
from nap import Response, SimPlatform
from nap.vendor import JTAG
from .jtag_peripheral_connector import JTAGPeripheralConnector


class TestPeripheral(Elaboratable):
    def range(self):
        return range(0, 1024)

    def __init__(self):
        self.mem = Memory(width=32, depth=1024, init=list(range(1024)))
        self.read_port = self.mem.read_port()
        self.write_port = self.mem.write_port()

        self.read_ctr = Signal(range(10))
        self.write_ctr = Signal(range(10))

    def elaborate(self, platform):
        m = Module()
        m.submodules.read_port = self.read_port
        m.submodules.write_port = self.write_port

        m.d.comb += Signal().eq(self.read_ctr)
        m.d.comb += Signal().eq(self.write_ctr)
        return m

    def handle_read(self, m, addr, data, read_done):
        with m.If(addr < 1024):
            m.d.sync += self.read_ctr.eq(self.read_ctr + 1)
            m.d.sync += self.read_port.addr.eq(addr)
            with m.If(self.read_ctr == 5):
                m.d.sync += data.eq(self.read_port.data)
                m.d.sync += self.read_ctr.eq(0)
                read_done(Response.OK)
        with m.Else():
            read_done(Response.ERR)

    def handle_write(self, m, addr, data, write_done):
        with m.If(addr < 1024):
            m.d.sync += self.write_ctr.eq(self.write_ctr + 1)
            m.d.sync += self.write_port.addr.eq(addr)
            with m.If(self.write_ctr == 5):
                m.d.comb += self.write_port.data.eq(data)
                m.d.comb += self.write_port.en.eq(1)
                m.d.sync += self.write_ctr.eq(0)
                write_done(Response.OK)
        with m.Else():
            write_done(Response.ERR)


class TestJTAGPeripheralConnectorFSM(unittest.TestCase):
    def check_read_write(self, tdi_delay=0, tdo_delay=0):
        platform = SimPlatform()

        jtag_device = JTAG()
        m = Module()
        test_peripheral = m.submodules.test_peripheral = TestPeripheral()
        m.submodules.jtag = JTAGPeripheralConnector(test_peripheral, jtag_device, jtag_domain="sync")

        jtag_testbench = JTAG()
        jtag_device.shift_tdi = jtag_testbench.shift_tdi
        jtag_device.shift_tdo = jtag_testbench.shift_tdi
        if tdi_delay == 0:
            jtag_device.tdi = jtag_testbench.tdi
        else:
            tdi_delay_chain = [jtag_device.tdi] + [Signal() for _ in range(tdi_delay - 1)] + [jtag_testbench.tdi]
            for src, dst in zip(tdi_delay_chain[1:], tdi_delay_chain[:-1]):
                m.d.sync += dst.eq(src)
        if tdo_delay == 0:
            jtag_device.tdo = jtag_testbench.tdo
        else:
            tdo_delay_chain = [jtag_testbench.tdo] + [Signal() for _ in range(tdo_delay - 1)] + [jtag_device.tdo]
            for src, dst in zip(tdo_delay_chain[1:], tdo_delay_chain[:-1]):
                m.d.sync += dst.eq(src)

        result = 0

        def shift_word(value, width=32):
            yield jtag_testbench.shift_tdi.eq(1)
            result_str = ""
            output_string = list(reversed("{{:0{}b}}".format(width).format(value)))
            for bit in output_string:
                if bit == "1":
                    yield jtag_testbench.tdi.eq(1)
                else:
                    yield jtag_testbench.tdi.eq(0)
                result_str += str((yield jtag_testbench.tdo))
                yield
            global result
            result = int("".join(reversed(result_str)), 2)
            yield jtag_testbench.shift_tdi.eq(0)

        def shift_bit(value):
            yield from shift_word(value, width=1)

        def testbench():
            fsm = platform.fragment.subfragments[1][0].find_generated("fsm")
            global result

            def assert_state(desired_state):
                if tdi_delay == 0 and tdo_delay == 0:
                    assert fsm.decoding[(yield fsm.state)] == desired_state

            def write(addr, value):
                for _ in range(10):
                    yield
                yield from shift_bit(0)  # wakeup
                yield from shift_bit(1)  # wakeup
                yield from assert_state("IDLE1")
                yield from shift_word(addr)  # address
                yield from assert_state("ADDR31")
                yield from shift_bit(1)  # write
                yield from assert_state("RW_CMD")
                yield from shift_word(value)
                yield from assert_state("WRITE31")
                # read wait
                timeout = 100
                for t in range(timeout):
                    if t == 0:
                        yield from assert_state("WRITE31")
                    else:
                        yield from assert_state("WRITE_WAIT")
                    yield from shift_bit(1)
                    if result == 1:
                        break
                    if t == timeout - 1:
                        raise TimeoutError()
                yield from assert_state("WRITE_STATUS")
                yield from shift_bit(0)  # read status
                assert result == 0
                yield from assert_state("IDLE0")

            def read(addr, expected_value):
                for _ in range(10):
                    yield
                yield from shift_bit(0)  # wakeup
                yield from shift_bit(1)  # wakeup
                yield from assert_state("IDLE1")
                yield from shift_word(addr)  # address
                yield from assert_state("ADDR31")
                yield from shift_bit(0)  # read
                yield from assert_state("RW_CMD")
                # read wait
                timeout = 100
                for t in range(timeout):
                    if t == 0:
                        yield from assert_state("RW_CMD")
                    else:
                        yield from assert_state("READ_WAIT")
                    yield from shift_bit(1)
                    if result == 1:
                        break
                    if t == timeout - 1:
                        raise TimeoutError()
                yield from assert_state("READ0")
                yield from shift_word(0)
                assert result == expected_value
                yield from assert_state("READ_STATUS")
                yield from shift_bit(0)  # read status
                assert result == 0
                yield from assert_state("IDLE0")

            for addr in range(0, 100, 10):
                yield from write(addr, addr + 1)
                yield from read(addr, addr + 1)

        platform.add_sim_clock("sync", 100e6)
        platform.sim(m, testbench)

    def test_read_transaction(self):
        self.check_read_write()

    def test_read_transaction_tdi_delay(self):
        self.check_read_write(tdi_delay=1, tdo_delay=0)

    def test_read_transaction_tdo_delay(self):
        self.check_read_write(tdi_delay=0, tdo_delay=1)

    def test_read_transaction_delay(self):
        self.check_read_write(tdi_delay=10, tdo_delay=10)



import lzma
import unittest
from os.path import dirname, join
from nmigen import Signal
from nap import SimPlatform
from .hispi_rx import LaneManager


class FakeSensorResource:
    def __init__(self):
        self.lvds = Signal(4)
        self.lvds_clk = Signal()


class TestHispi(unittest.TestCase):
    def check_hispi_lane_manager(self, datafile):
        platform = SimPlatform()
        input_data = Signal(12)
        dut = LaneManager(input_data)

        def testbench():
            def lane_n_generator(n):
                with lzma.open(join(dirname(__file__), datafile), "r") as f:
                    for line in f:
                        line = line.replace(b" ", b"")
                        for i in range(12):
                            val = "0" if line[i + (n * 12)] == ord("0") else "1"
                            yield val

            generator = lane_n_generator(0)
            last_valid = 0
            line_ctr = 0
            for i in range(200000):
                try:
                    if (yield dut.do_bitslip):
                        next(generator)
                    word = int("".join(reversed([next(generator) for _ in range(12)])),
                               2)  # TODO: figure out, how this reversed() affects the real world
                    yield input_data.eq(word)
                except RuntimeError:  # this is raised when we are done instead of StopIteration
                    break

                if (last_valid == 0) and ((yield dut.output.valid) == 1):
                    last_valid = 1
                    line_ctr = 0
                elif (last_valid == 1) and ((yield dut.output.valid) == 1):
                    line_ctr += 1
                elif (last_valid == 1) and ((yield dut.output.valid) == 0):
                    last_valid = 0
                    assert (line_ctr + 1) * 4 == 2304

                yield
            assert (yield dut.is_aligned) == 1, "dut is not aligned"

        platform.add_sim_clock("sync", 100e6)
        platform.sim(dut, testbench)

    def test_hispi_lane_manager_old_data(self):
        self.check_hispi_lane_manager("test_data_old.txt.lzma")

    def test_hispi_lane_manager_new_data(self):
        self.check_hispi_lane_manager("test_data_new.txt.lzma")


from random import Random
from nap import write_to_stream, read_from_stream, do_nothing

__all__ = ["write_frame_to_stream", "read_frame_from_stream", "to_8bit_rgb", "crop"]


def write_frame_to_stream(stream, frame, timeout=100, pause=False):
    random = Random(0)
    for y, line in enumerate(frame):
        for x, px in enumerate(line):
            if (random.random() < 0.3) and pause:
                yield from do_nothing()
            yield from write_to_stream(
                stream, payload=int(px),
                timeout=timeout,
                line_last=(x == (len(line) - 1)),
                frame_last=(y == (len(frame) - 1)) & (x == (len(line) - 1)),
            )
        if (random.random() < 0.3) and pause:
            yield from do_nothing()


def read_frame_from_stream(stream, timeout=100, pause=False):
    random = Random(1)
    frame = [[]]
    while True:
        if (random.random() < 0.3) and pause:
            yield from do_nothing()
        px, line_last, frame_last = (yield from read_from_stream(stream, timeout=timeout,
                                                                 extract=("payload", "line_last", "frame_last")))
        frame[-1].append(px)
        if frame_last:
            return frame
        if line_last:
            frame.append([])


def to_8bit_rgb(image_24bit):
    return [
        [[px & 0xff, (px >> 8) & 0xff, (px >> 16) & 0xff] for px in line]
        for line in image_24bit
    ]


def crop(frame, left=0, right=0, top=0, bottom=0):
    return [line[left:len(line)-right] for line in frame[top:len(frame)-bottom]]



# reports the clock frequencies of the design using csr infrastructure

from nmigen import *
from nap import StatusSignal, driver_property

__all__ = ["ClockingDebug", "ClockDebug"]


class ClockingDebug(Elaboratable):
    def __init__(self, *args):
        self.clockdomains = args

    def elaborate(self, platform):
        m = Module()

        for cd in self.clockdomains:
            m.submodules[cd] = ClockDebug(cd)

        return m


class ClockDebug(Elaboratable):
    def __init__(self, domain_name, reset_less=False):
        self.domain_name = domain_name
        self.reset_less = reset_less

        self.counter = StatusSignal(32)
        self.is_reset = StatusSignal()

    def elaborate(self, platform):
        m = Module()

        m.d[self.domain_name] += self.counter.eq(self.counter + 1)
        m.d.comb += self.is_reset.eq(ResetSignal(self.domain_name))

        return m

    @driver_property
    def mhz(self):
        from time import sleep, time
        initial_counter = self.counter
        start = time()
        sleep(0.1)
        counter_difference = (self.counter - initial_counter)
        return counter_difference * (1 / (time() - start)) / 1e6



import struct
import sys
from io import BytesIO


def flip32(data):
    sl = struct.Struct('<I')
    sb = struct.Struct('>I')
    d = bytearray(len(data))
    for offset in range(0, len(data), 4):
        sb.pack_into(d, offset, sl.unpack_from(data, offset)[0])
    return d


def bit2bin(bitstream, flip_data=True):
    bitfile = BytesIO(bitstream)

    short = struct.Struct('>H')
    ulong = struct.Struct('>I')

    l = short.unpack(bitfile.read(2))[0]
    if l != 9:
        raise Exception("Missing <0009> header (0x%x), not a bit file" % l)
    bitfile.read(l)
    l = short.unpack(bitfile.read(2))[0]
    d = bitfile.read(l)
    if d != b'a':
        raise Exception("Missing <a> header, not a bit file")

    l = short.unpack(bitfile.read(2))[0]
    d = bitfile.read(l)

    KEYNAMES = {b'b': "Partname", b'c': "Date", b'd': "Time"}
    while 1:
        k = bitfile.read(1)
        if not k:
            bitfile.close()
            raise Exception("unexpected EOF")
        elif k == b'e':
            l = ulong.unpack(bitfile.read(4))[0]
            d = bitfile.read(l)
            if flip_data:
                return flip32(d)
            else:
                return d
        elif k in KEYNAMES:
            l = short.unpack(bitfile.read(2))[0]
            d = bitfile.read(l)
        else:
            print("Unexpected key: %s" % k)
            l = short.unpack(bitfile.read(2))[0]
            d = bitfile.read(l)


if __name__ == "__main__":
    input_file_name = sys.argv[1]
    if input_file_name == "-":
        input_file = sys.stdin
    else:
        input_file = open(input_file_name, 'rb')
    input_bitstream = input_file.read()
    sys.stdout.buffer.write(bit2bin(input_bitstream))



from nmigen import *
from nmigen.lib.fifo import SyncFIFOBuffered, SyncFIFO, AsyncFIFOBuffered, AsyncFIFO
from nap import Stream, StatusSignal

__all__ = ["BufferedAsyncStreamFIFO", "UnbufferedAsyncStreamFIFO", "BufferedSyncStreamFIFO", "UnbufferedSyncStreamFIFO"]


class StreamFIFO(Elaboratable):
    def __init__(self, input: Stream, fifo_type, output_stream_name="fifo_out", **fifo_args):
        self.input = input
        self.output = input.clone(name=output_stream_name)
        if "r_domain" in fifo_args:
            self.output_domain = fifo_args["r_domain"]
        self.fifo = fifo_type(width=len(Cat(self.input.payload_signals.values())), **fifo_args)
        self.depth = fifo_args['depth']

        self.r_level = StatusSignal(range(self.fifo.depth + 1))
        self.w_level = StatusSignal(range(self.fifo.depth + 1))

    def elaborate(self, platform):
        m = Module()
        fifo = m.submodules.fifo = self.fifo

        if self.depth == 0:
            m.d.comb += self.output.connect_upstream(self.input)
        else:
            m.d.comb += self.r_level.eq(fifo.r_level)
            m.d.comb += self.w_level.eq(fifo.w_level)

            m.d.comb += self.input.ready.eq(fifo.w_rdy)
            m.d.comb += fifo.w_data.eq(Cat(self.input.payload_signals.values()))
            m.d.comb += fifo.w_en.eq(self.input.valid)

            m.d.comb += Cat(self.output.payload_signals.values()).eq(fifo.r_data)
            m.d.comb += self.output.valid.eq(fifo.r_rdy)
            m.d.comb += fifo.r_en.eq(self.output.ready)

        return m


def BufferedSyncStreamFIFO(input: Stream, depth, **kwargs):
    return StreamFIFO(input, SyncFIFO, depth=depth, **kwargs)


def UnbufferedSyncStreamFIFO(input: Stream, depth, **kwargs):
    return StreamFIFO(input, SyncFIFOBuffered, depth=depth, **kwargs)


def BufferedAsyncStreamFIFO(input, depth, i_domain="sync", o_domain="sync", exact_depth=False, **kwargs):
    return StreamFIFO(
        input, AsyncFIFOBuffered, depth=depth, r_domain=o_domain, w_domain=i_domain, exact_depth=exact_depth,
        **kwargs
    )


def UnbufferedAsyncStreamFIFO(input, depth, i_domain="sync", o_domain="sync", exact_depth=False, **kwargs):
    return StreamFIFO(
        input, AsyncFIFO, depth=depth, r_domain=o_domain, w_domain=i_domain, exact_depth=exact_depth,
        **kwargs
    )



import inspect
import subprocess
import textwrap
from pathlib import Path

from nmigen import Fragment
from nmigen._toolchain import require_tool
from nmigen.back import rtlil
from shutil import rmtree


def assert_formal(spec, mode="bmc", depth=1):
    functions = []
    test_class = None
    caller_path = ""
    stack = inspect.stack()
    for frame in stack[1:]:
        if "unittest" in frame.filename:
            filename = "__".join(reversed(functions))
            if test_class:
                filename = f'{test_class}__{filename}'
            break
        functions.append(frame.function)
        try:
            test_class = frame.frame.f_locals['self'].__class__.__name__
        except:
            pass
        caller_path = frame.filename

    target_dir = Path(caller_path).parent / ".sim_results"
    target_dir.mkdir(exist_ok=True)
    if (target_dir / filename).exists():
        rmtree(target_dir / filename)

    if mode == "hybrid":
        # A mix of BMC and k-induction, as per personal communication with Claire Wolf.
        script = "setattr -unset init w:* a:nmigen.sample_reg %d"
        mode = "bmc"
    else:
        script = ""

    config = textwrap.dedent("""\
        [options]
        mode {mode}
        depth {depth}
        wait on
        [engines]
        smtbmc
        [script]
        read_ilang top.il
        prep
        {script}
        [file top.il]
        {rtlil}
    """).format(
        mode=mode,
        depth=depth,
        script=script,
        rtlil=rtlil.convert(Fragment.get(spec, platform="formal"))
    )
    with subprocess.Popen([require_tool("sby"), "-f", "-d", filename], cwd=str(target_dir),
                          universal_newlines=True,
                          stdin=subprocess.PIPE, stdout=subprocess.PIPE) as proc:
        stdout, stderr = proc.communicate(config)
        if proc.returncode != 0:

            assert False, "Formal verification failed:\n" + stdout + "\n\n" + f"vcd: {str(target_dir / filename)}/engine_0/trace.vcd"


import unittest
from collections import defaultdict
from bitarray import bitarray
from nmigen import *
from nap import SimPlatform, PacketizedStream, write_to_stream, read_from_stream, LegalStreamSource, verify_stream_output_contract
from .bit_stuffing import BitStuffer
from .huffman_encoder import HuffmanEncoder


class HuffmanTest(unittest.TestCase):
    def test_hello_world(self):
        platform = SimPlatform()
        m = Module()

        input = PacketizedStream(8)
        input_data = "hello, world :)"
        distribution = defaultdict(lambda: 0)
        for c in input_data:
            distribution[ord(c)] += 1
        huffman = m.submodules.huffman = HuffmanEncoder(input, distribution)

        def write_process():
            for i, c in enumerate(input_data):
                yield from write_to_stream(input, payload=ord(c), last=(i == len(input_data) - 1))

        def read_process():
            read = ""
            while True:
                data, length, last = (yield from read_from_stream(huffman.output, extract=("payload", "current_width", "last")))
                bitstring = "{:0255b}".format(data)[::-1][:length]
                read += bitstring
                if last:
                    break
            print(read)
            decode_iter = bitarray(read).iterdecode({k: bitarray(v[::-1]) for k, v in huffman.table.items()})
            decoded = ""
            try:
                for c in decode_iter:
                    decoded += chr(c)
            except ValueError:  # Decoding may not finish with the byte boundary
                pass
            self.assertEqual(input_data, decoded)

        platform.add_sim_clock("sync", 100e6)
        platform.add_process(write_process, "sync")
        platform.sim(m, read_process)

    def test_hello_world_bit_stuffing(self):
        platform = SimPlatform()
        m = Module()

        input = PacketizedStream(8)
        input_data = "hello, world :)"
        distribution = defaultdict(lambda: 0)
        for c in input_data:
            distribution[ord(c)] += 1
        huffman = m.submodules.huffman = HuffmanEncoder(input, distribution)
        bit_stuffing = m.submodules.bit_stuffing = BitStuffer(huffman.output, 8)

        def write_process():
            for i, c in enumerate(input_data):
                yield from write_to_stream(input, payload=ord(c), last=(i == len(input_data) - 1))

        def read_process():
            read = []
            while True:
                payload, last = (yield from read_from_stream(bit_stuffing.output, extract=("payload", "last")))
                read.append("{:08b}".format(payload))
                if last:
                    break
            read_bitarray = "".join(x[::-1] for x in read)
            print(read_bitarray)
            decode_iter = bitarray(read_bitarray).iterdecode({k: bitarray(v[::-1]) for k, v in huffman.table.items()})
            for c, expected in zip(decode_iter, input_data):
                self.assertEquals(chr(c), expected)

        platform.add_sim_clock("sync", 100e6)
        platform.add_process(write_process, "sync")
        platform.sim(m, read_process)

    def test_output_stream_properties(self):
        input = PacketizedStream(8)
        verify_stream_output_contract(HuffmanEncoder(input, {i: i for i in range(256)}), support_modules=(LegalStreamSource(input),))



import unittest
from nmigen import *
from nap import SimPlatform
from nap.stream import PacketizedStream, verify_stream_output_contract, LegalStreamSource, write_to_stream, read_from_stream
from . import ZeroRleEncoder, RleEncodingSpace


class RleTest(unittest.TestCase):
    def test_basic(self):
        platform = SimPlatform()
        m = Module()

        input = PacketizedStream(8)
        input_data = [1, 0, 1, *([0] * 14), 1]
        run_length_options = [3, 10, 27, 80, 160]

        rle = m.submodules.rle = ZeroRleEncoder(input, RleEncodingSpace(range(0, 255), run_length_options, zero_value=0))

        def write_process():
            for x in input_data:
                yield from write_to_stream(input, payload=x)

        def read_process():
            received = []
            while True:
                try:
                    received.append((yield from read_from_stream(rle.output)))
                except TimeoutError:
                    break
            decoded = []
            for x in received:
                if x < 256:
                    decoded.append(x)
                else:
                    decoded += ([0] * run_length_options[x - 256])
            self.assertEquals(input_data, decoded)

        platform.add_sim_clock("sync", 100e6)
        platform.add_process(write_process, "sync")
        platform.sim(m, read_process)

    def test_output_stream_properties(self):
        input = PacketizedStream(8)
        encoding_space = RleEncodingSpace(range(0, 255), [3, 10, 27, 80, 160], zero_value=0)
        verify_stream_output_contract(ZeroRleEncoder(input, encoding_space), support_modules=(LegalStreamSource(input),))



import numpy as np


def interleave(*args):
    n = len(args)
    w, h = args[0].shape
    a = np.zeros((w * n, h))
    for i, arg in enumerate(args):
        a[i::n] = arg
    return a


def full_width(w, stages):
    return int(width_factor(stages) * w)


def width_factor(stages, factor=1.0):
    if stages == 1:
        return 2 * factor
    else:
        return factor * 3 / 2 + width_factor(stages - 1, factor / 2)


def real_width(encoded_width, stages):
    factor = width_factor(stages)
    return int(encoded_width // factor)


def pack(image, levels):
    h, w = image.shape
    result = np.zeros((h, full_width(w, levels)), dtype=image.dtype)
    orig_result = result

    for level in reversed(range(levels)):
        hf1 = image[:h // 2, w // 2:]
        hf2 = image[h // 2:, :w // 2]
        hf3 = image[h // 2:, w // 2:]

        result[:h // 2, -w * 3 // 2:-w * 2 // 2] = hf1
        result[:h // 2, -w * 2 // 2:-w * 1 // 2] = hf2
        result[:h // 2, -w * 1 // 2:] = hf3

        if level == 0:
            result[:h // 2, :w // 2] = image[:h // 2, :w // 2]
        else:
            result = result[5::2, :-w * 3 // 2]
            image = image[:h // 2, :w // 2]
            w //= 2
            h //= 2

    return orig_result


def unpack(image, levels):
    h, w = image.shape
    w = real_width(w, levels)
    result = np.zeros((h, w), dtype=image.dtype)
    orig_result = result

    for level in reversed(range(levels)):
        hf1 = image[:h // 2, -w * 3 // 2: -w * 2 // 2]
        hf2 = image[:h // 2, -w * 2 // 2: -w * 1 // 2]
        hf3 = image[:h // 2, -w * 1 // 2:]
        result[:h // 2, w // 2:] = hf1
        result[h // 2:, :w // 2] = hf2
        result[h // 2:, w // 2:] = hf3

        if level == 0:
            result[:h // 2, :w // 2] = image[:h // 2, :w // 2]
        else:
            result = result[:h // 2, :w // 2]
            image = image[5::2, :-w * 3 // 2]
            w //= 2
            h //= 2

    return orig_result



from nmigen import *
from ..data_structure import DOWNWARDS
from .stream import BasicStream

__all__ = ["PacketizedFirstStream"]


class PacketizedFirstStream(BasicStream):
    """
    A stream that carries a payload and can separate Packets via a first signal that is asserted on the
    first word of a packet
    """

    def __init__(self, payload_shape, name=None, src_loc_at=1):
        super().__init__(payload_shape, name, src_loc_at=1 + src_loc_at)
        self.first = Signal() @ DOWNWARDS



# Experimental gateware for the usb3 plugin module
from nmigen import *
from nap import *


class Top(Elaboratable):
    def elaborate(self, platform):
        m = Module()

        clocking = m.submodules.clocking = ClockingDebug("sync", "sync_in", "ft601")

        plugin = platform.request("plugin_stream_input")
        rx = m.submodules.rx = PluginModuleStreamerRx(plugin, domain_name="sync")

        ft601 = platform.request("ft601")
        m.submodules.ft601 = FT601StreamSink(ft601, rx.output, domain_name="ft601")

        return m


if __name__ == "__main__":
    cli(Top, runs_on=(Usb3PluginPlatform,), possible_socs=(JTAGSocPlatform,))



# An experiment that allows to capture the raw mipi data with 1:8 serdes to feed into the
# simulation tests

from nmigen import *
from nmigen.build import Subsignal, Pins, PinsN, Attrs, DiffPairs, Resource
from nap import *
from nap.vendor.xilinx_s7 import IDelayCtrl


class Top(Elaboratable):
    def __init__(self):
        self.sensor_reset_n = ControlSignal(name='sensor_reset', reset=1)
        self.enable_write = ControlSignal()
        self.change_packet = ControlSignal()

    def elaborate(self, platform):
        m = Module()

        platform.ps7.fck_domain(100e6, "axi_hp")

        # Control Pane
        i2c_pads = platform.request("i2c")
        m.submodules.i2c = BitbangI2c(i2c_pads)

        clocking = m.submodules.clocking = ClockingDebug("sync", "sensor_clk", "axi_hp")

        # Input Pipeline
        platform.add_resources([
            Resource("sensor2", 0,
                     Subsignal("shutter", Pins("25", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     Subsignal("trigger", Pins("27", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     Subsignal("reset", PinsN("31", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     Subsignal("clk", Pins("33", dir='o', conn=("expansion", 0)), Attrs(IOSTANDARD="LVCMOS25")),
                     Subsignal("lvds_clk", DiffPairs("51", "53", dir='i', conn=("expansion", 0)), Attrs(IOSTANDARD="LVDS_25", DIFF_TERM="TRUE")),
                     Subsignal("lvds", DiffPairs("41 45 55 65", "43 47 57 67", dir='i', conn=("expansion", 0)), Attrs(IOSTANDARD="LVDS_25", DIFF_TERM="TRUE")),
                     ),
        ])
        sensor = platform.request("sensor2")
        platform.ps7.fck_domain(24e6, "sensor_clk")
        m.d.comb += sensor.clk.eq(ClockSignal("sensor_clk"))
        m.d.comb += sensor.reset.eq(~self.sensor_reset_n)

        # this experiment requires bodging the MicroR2 in a way to connect the mipi lanes of the ar0330
        # to the pins where normally the hispi would be connected. the normal hispi traces must be cut.
        mipi_clock = sensor.lvds[0]
        mipi_lane4 = sensor.lvds[1]
        mipi_lane3 = sensor.lvds_clk
        mipi_lane2 = sensor.lvds[2]
        mipi_lane1 = sensor.lvds[3]

        m.domains.sync = ClockDomain()
        m.submodules.clock_rx = MipiClockRxPhy(mipi_clock, ddr_domain="ddr")
        lanes = [mipi_lane1, mipi_lane2, mipi_lane3, mipi_lane4]
        lane_phys = [MipiLaneRxPhy(p, ddr_domain="ddr") for p in lanes]
        for i, phy in enumerate(lane_phys):
            m.submodules[f'lane{i + 2}_phy'] = phy

        platform.ps7.fck_domain(200e6, "idelay_ref")
        m.submodules.idelay_ctrl = IDelayCtrl("idelay_ref")

        data_stream = PacketizedStream(32)
        m.d.comb += data_stream.payload.eq(Cat(phy.output for phy in lane_phys))
        m.d.comb += data_stream.valid.eq(self.enable_write)
        m.d.comb += data_stream.last.eq(self.change_packet)

        p = Pipeline(m)
        p += BufferedAsyncStreamFIFO(data_stream, 2048, o_domain="axi_hp")
        p += DramPacketRingbufferStreamWriter(p.output, max_packet_size=0x800000, n_buffers=1)

        return m

    @driver_method
    def kick_sensor(self):
        from os import system
        system("cat /axiom-api/scripts/kick/value")


if __name__ == "__main__":
    cli(Top, runs_on=(MicroR2Platform,), possible_socs=(ZynqSocPlatform,))



import unittest
import random
from nmigen import *
from nmigen.hdl.ast import Assert, Initial, Assume
from nap import SimPlatform, PacketizedStream, BasicStream, assert_formal, write_packet_to_stream, read_packet_from_stream, BufferedSyncStreamFIFO
from nap.stream.formal_util import verify_stream_output_contract, LegalStreamSource
from . import LastWrapper, GenericMetadataWrapper


class LastWrapperContract(Elaboratable):
    def __init__(self, last_wrapper, packets=(1, 27, 3, 5)):
        self.last_wrapper = last_wrapper
        self.last_data = Array((i == l - 1) for l in packets for i in range(l))
        self.n_packets = len(packets)

    def elaborate(self, platform):
        m = Module()

        m.submodules.last_wrapper = self.last_wrapper

        lw_input = self.last_wrapper.input
        lw_output = self.last_wrapper.output
        with m.If(Initial()):
            m.d.comb += Assume(ResetSignal())

        write_ctr = Signal(range(len(self.last_data)))
        with m.If(write_ctr < len(self.last_data) - 1):
            m.d.comb += lw_input.valid.eq(1)
            m.d.comb += lw_input.last.eq(self.last_data[write_ctr])
            with m.If(lw_input.ready):
                m.d.sync += write_ctr.eq(write_ctr + 1)

        read_ctr = Signal.like(write_ctr)
        with m.If(read_ctr < len(self.last_data) - 1):
            m.d.comb += lw_output.ready.eq(1)
            with m.If(lw_output.valid):
                m.d.sync += read_ctr.eq(read_ctr + 1)
                with m.If(~Initial()):
                    m.d.comb += Assert(lw_output.last == self.last_data[read_ctr])

        return m


class LastWrapperTest(unittest.TestCase):
    def test_randomized(self):
        platform = SimPlatform()

        input_stream = PacketizedStream(32)
        dut = LastWrapper(input_stream, lambda i: BufferedSyncStreamFIFO(i, 10), last_rle_bits=4)

        random.seed(0)
        test_packets = [
            [random.randint(0, 2**32) for _ in range(12)],
            [random.randint(0, 2**32) for _ in range(24)],
            [random.randint(0, 2**32) for _ in range(1)],
            [random.randint(0, 2**32) for _ in range(1000)],
            [random.randint(0, 2**32) for _ in range(1000)],
            [random.randint(0, 2 ** 32) for _ in range(12)],
            [random.randint(0, 2 ** 32) for _ in range(1000)],
        ]

        def writer_process():
            for packet in test_packets:
                yield from write_packet_to_stream(input_stream, packet)
        platform.add_process(writer_process, "sync")

        def reader_process():
            read_packets = []
            while len(read_packets) < len(test_packets):
                read = (yield from read_packet_from_stream(dut.output))
                read_packets.append(read)
                print([len(p) for p in read_packets])

            self.assertEqual(read_packets, test_packets)
        platform.add_process(reader_process, "sync")

        platform.add_sim_clock("sync", 100e6)
        platform.sim(dut)

    def test_long_fifo(self):
        platform = SimPlatform()

        input_stream = PacketizedStream(32)
        dut = LastWrapper(input_stream, lambda i: BufferedSyncStreamFIFO(i, 200), last_rle_bits=10)

        random.seed(0)
        test_packets = [
            [random.randint(0, 2**32) for _ in range(12)],
            [random.randint(0, 2**32) for _ in range(24)],
            [random.randint(0, 2**32) for _ in range(1)],
            [random.randint(0, 2**32) for _ in range(1000)],
            [random.randint(0, 2**32) for _ in range(1000)],
            [random.randint(0, 2 ** 32) for _ in range(12)],
            [random.randint(0, 2 ** 32) for _ in range(1000)],
        ]

        def writer_process():
            for packet in test_packets:
                yield from write_packet_to_stream(input_stream, packet)
        platform.add_process(writer_process, "sync")

        def reader_process():
            read_packets = []
            while len(read_packets) < len(test_packets):
                read = (yield from read_packet_from_stream(dut.output))
                read_packets.append(read)
                print([len(p) for p in read_packets])

            self.assertEqual(read_packets, test_packets)
        platform.add_process(reader_process, "sync")

        platform.add_sim_clock("sync", 100e6)
        platform.sim(dut)

    def test_last_wrapper_contract(self):
        dut = LastWrapperContract(LastWrapper(PacketizedStream(32), lambda i: BufferedSyncStreamFIFO(i, 10)))
        assert_formal(dut, mode="hybrid", depth=10)

    def test_output_stream_contract(self):
        input_stream = PacketizedStream(32)
        dut = LastWrapper(input_stream, lambda i: BufferedSyncStreamFIFO(i, 10), last_rle_bits=3)
        verify_stream_output_contract(dut, support_modules=(LegalStreamSource(input_stream),))

    def test_core_output_stream_contract(self):
        input_stream = PacketizedStream(32)
        device_input_stream: BasicStream
        def core_producer(i):
            nonlocal device_input_stream
            device_input_stream = i
            return LegalStreamSource(i.clone())
        dut = LastWrapper(input_stream, core_producer, last_rle_bits=3)
        verify_stream_output_contract(dut, stream_output=device_input_stream, support_modules=(LegalStreamSource(input_stream),))


class GenericMetadataWrapperTest(unittest.TestCase):
    def test_randomized(self):
        platform = SimPlatform()

        input_stream = PacketizedStream(32)
        dut = GenericMetadataWrapper(input_stream, lambda i: BufferedSyncStreamFIFO(i, 10, output_stream_name="core_output"), fifo_depth=11)

        random.seed(0)
        test_packets = [
            [random.randint(0, 2**32) for _ in range(12)],
            [random.randint(0, 2**32) for _ in range(24)],
            [random.randint(0, 2**32) for _ in range(1)],
            [random.randint(0, 2**32) for _ in range(1000)],
            [random.randint(0, 2**32) for _ in range(1000)],
            [random.randint(0, 2 ** 32) for _ in range(12)],
            [random.randint(0, 2 ** 32) for _ in range(1000)],
        ]

        def writer_process():
            for packet in test_packets:
                yield from write_packet_to_stream(input_stream, packet)
        platform.add_process(writer_process, "sync")

        def reader_process():
            read_packets = []
            while len(read_packets) < len(test_packets):
                read = (yield from read_packet_from_stream(dut.output))
                read_packets.append(read)
                print([len(p) for p in read_packets])

            self.assertEqual(read_packets, test_packets)
        platform.add_process(reader_process, "sync")

        platform.add_sim_clock("sync", 100e6)
        platform.sim(dut)

    def test_output_stream_contract(self):
        input_stream = PacketizedStream(32)
        dut = GenericMetadataWrapper(input_stream, lambda i: BufferedSyncStreamFIFO(i, 10))
        verify_stream_output_contract(dut, support_modules=(LegalStreamSource(input_stream),))

    def test_core_output_stream_contract(self):
        input_stream = PacketizedStream(32)
        device_input_stream: BasicStream
        def core_producer(i):
            nonlocal device_input_stream
            device_input_stream = i
            return LegalStreamSource(i.clone())
        dut = GenericMetadataWrapper(input_stream, core_producer)
        verify_stream_output_contract(dut, stream_output=device_input_stream, support_modules=(LegalStreamSource(input_stream),))



from .peripheral import Response
from nap.util.nmigen_misc import iterator_with_if_elif

__all__ = ["PeripheralsAggregator"]


class PeripheralsAggregator:
    def __init__(self):
        """
        A helper class that behaves like a Peripheral but proxies its read/write request to downstream peripherals
        based on their memorymap.
        """
        self.downstream_peripherals = []

    def add_peripheral(self, peripheral):
        assert callable(peripheral.handle_read) and callable(peripheral.handle_write)
        assert isinstance(peripheral.range(), range)
        self.downstream_peripherals.append(peripheral)

    def range(self):
        return range(
            min(p.range().start for p in self.downstream_peripherals),
            max(p.range().stop for p in self.downstream_peripherals)
        )

    def handle_read(self, m, addr, data, read_done_callback):
        for cond, peripheral in iterator_with_if_elif(self.downstream_peripherals, m):
            address_range = peripheral.memorymap.absolute_range_of_direct_children.range()
            translated_address_range = range(
                address_range.start - self.range().start,
                address_range.stop - self.range().start,
            )
            with cond((addr >= translated_address_range.start) & (addr < translated_address_range.stop)):
                peripheral.handle_read(m, addr - translated_address_range.start, data, read_done_callback)
        if self.downstream_peripherals:
            with m.Else():
                read_done_callback(Response.ERR)
        else:
            read_done_callback(Response.ERR)

    def handle_write(self, m, addr, data, write_done_callback):
        for cond, peripheral in iterator_with_if_elif(self.downstream_peripherals, m):
            address_range = peripheral.memorymap.absolute_range_of_direct_children.range()
            translated_address_range = range(
                address_range.start - self.range().start,
                address_range.stop - self.range().start,
            )
            with cond((addr >= translated_address_range.start) & (addr < translated_address_range.stop)):
                peripheral.handle_write(m, addr - translated_address_range.start, data, write_done_callback)
        if self.downstream_peripherals:
            with m.Else():
                write_done_callback(Response.ERR)
        else:
            write_done_callback(Response.ERR)



from nmigen import *
from nmigen.hdl.ast import Cover, Assume, Initial, Assert

from .stream import Stream
from nap.util.formal import assert_formal

__all__ = ["verify_stream_output_contract", "LegalStreamSource"]

class StreamOutputCoverSpec(Elaboratable):
    """
    the valid signal MUST NOT depend on the ready signal.
    the other way round a dependency is okay.
    """
    def __init__(self, stream_output: Stream):
        self.stream_output = stream_output

    def elaborate(self, platform):
        m = Module()

        with m.If(Initial()):
            m.d.comb += Assume(ResetSignal())
        with m.Else():
            m.d.comb += Assume(~ResetSignal())
            m.d.comb += Assume(~self.stream_output.ready)
            m.d.comb += Cover(self.stream_output.valid)

        return m


def verify_stream_output_contract_cover(module, stream_output, support_modules=()):
    m = Module()
    for s in support_modules:
        m.submodules += s
    m.submodules.dut = module
    m.submodules.spec = StreamOutputCoverSpec(stream_output)
    assert_formal(m, mode="cover", depth=10)


class StreamOutputAssertSpec(Elaboratable):
    """
    Assert that the payload signals of a stream do not change if valid is pulled high and the transaction wasnt accepted
    by the next node by pulling ready high.
    """
    def __init__(self, stream_output: Stream):
        self.stream_output = stream_output

    def elaborate(self, platform):
        m = Module()

        with m.If(Initial()):
            m.d.comb += Assume(ResetSignal())
        with m.Else():
            m.d.comb += Assume(~ResetSignal())
            unfinished_transaction = Signal()
            with m.If(self.stream_output.ready & self.stream_output.valid):
                m.d.sync += unfinished_transaction.eq(0)
            with m.Elif(self.stream_output.valid):
                m.d.sync += unfinished_transaction.eq(1)
            m.d.comb += Assert(unfinished_transaction.implies(self.stream_output.valid))

            last_payload_signals = [Signal.like(s, name=f"{name}_last") for name, s in self.stream_output.payload_signals.items()]
            for l, s in zip(last_payload_signals, self.stream_output.payload_signals.values()):
                m.d.sync += l.eq(s)

            with m.If(unfinished_transaction):
                for l, s in zip(last_payload_signals, self.stream_output.payload_signals.values()):
                    m.d.comb += Assert(l == s)

        return m


def verify_stream_output_contract_assert(module, stream_output, support_modules=()):
    m = Module()
    for s in support_modules:
        m.submodules += s
    m.submodules.dut = module
    m.submodules.spec = StreamOutputAssertSpec(stream_output)
    assert_formal(m, mode="hybrid", depth=10)


class LegalStreamSource(Elaboratable):
    """A stream source that can be used to constrain stream input of cores"""
    def __init__(self, stream: Stream):
        self.output = stream

    def elaborate(self, platform):
        m = Module()

        unfinished_transaction = Signal()
        last_payload_signals = [Signal.like(s, name=f"{name}_last") for name, s in self.output.payload_signals.items()]
        with m.If(self.output.ready & self.output.valid):
            m.d.sync += unfinished_transaction.eq(0)
        with m.Elif(self.output.valid):
            m.d.sync += unfinished_transaction.eq(1)
            with m.If(~unfinished_transaction):
                for l, s in zip(last_payload_signals, self.output.payload_signals.values()):
                    m.d.sync += l.eq(s)

        with m.If(unfinished_transaction):
            m.d.comb += Assume(self.output.valid)
            for l, s in zip(last_payload_signals, self.output.payload_signals.values()):
                m.d.comb += Assume(l == s)

        return m


def verify_stream_output_contract(module, stream_output=None, support_modules=()):
    if stream_output is None:
        stream_output = module.output

    print("testing that valid does not depend on ready...")
    verify_stream_output_contract_cover(module, stream_output, support_modules)
    print("testing hold unacknowledged transactions...")
    verify_stream_output_contract_assert(module, stream_output, support_modules)



import re
import subprocess
import unittest
from glob import glob
from os.path import dirname


class TestBuilds(unittest.TestCase):
    pass


for file in glob("{}/*.py".format(dirname(__file__))):
    if "builds_test" in file:  # we are not executing ourselves
        continue
    try:
        subprocess.check_output(['python', file], stderr=subprocess.PIPE)
        continue
    except subprocess.CalledProcessError as e:
        usage = e.stderr.decode("utf-8")
    try:
        devices = re.search("-d\\W*{(.*?)}", usage).group(1).split(",")
        socs = re.search("-s\\W*{(.*?)}", usage).group(1).split(",")
    except Exception as e:
        print(e)
        print(file, usage)
        raise e

    for device in devices:
        for soc in socs:
            def make_test_builds(file, device):
                soc_local = str(soc)
                def test_builds(self):
                    command = ['python', file, '-e', '-d', device, '-s', soc_local]
                    print("running '{}'".format(' '.join(command)))
                    process = subprocess.Popen(command, stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE, stdin=subprocess.PIPE)
                    stdout, stderr = process.communicate()
                    print(stdout.decode("utf-8"))
                    print(stderr.decode("utf-8"))

                    self.assertEqual(0, process.returncode)

                return test_builds


            filename_stripped = file.split("/")[-1].replace(".py", "")
            setattr(TestBuilds, "test_{}_{}_{}".format(filename_stripped, device, soc), make_test_builds(file, device))



import subprocess
from shlex import quote
import os

from nmigen.build.run import BuildProducts

__all__ = ["run_ssh", "copy_ssh", "program_bitstream_ssh"]


default_host = os.getenv("SSH_HOST", "10.42.0.1")
default_user = os.getenv("SSH_USER", "operator")
default_password = os.getenv("SSH_PASSWORD", "axiom")


def run_ssh(cmd, host=default_host, user=default_user, password=default_password, sudo=False, sshpass=True):
    if sudo:
        cmd = "echo {} | sudo -S bash -c {}".format(password, quote(cmd))
    ssh_cmd = "ssh {}@{} {}".format(user, host, quote(cmd))
    if sshpass:
        ssh_cmd = "sshpass -p{} {}".format(password, ssh_cmd)
    print("\nexecuting: ", ssh_cmd)
    return subprocess.check_output(ssh_cmd, shell=True)


def copy_ssh(source, destination, host=default_host, user=default_user, password=default_password, sshpass=True):
    scp_cmd = "scp {} {}".format(quote(source), quote("{}@{}:{}".format(user, host, destination)))
    if sshpass:
        scp_cmd = "sshpass -p{} {}".format(password, scp_cmd)
    print("\nexecuting: ", scp_cmd)
    return subprocess.check_output(scp_cmd, shell=True)


def program_bitstream_ssh(platform, build_products: BuildProducts, name, **kwargs):
    fatbitstream_name = "{}.fatbitstream.sh".format(name)
    with build_products.extract(fatbitstream_name) as fatbitstream_file:
        copy_ssh(fatbitstream_file, "~/{}".format(fatbitstream_name))
    run_ssh("chmod +x {}".format(fatbitstream_name))
    run_ssh("./{}".format(fatbitstream_name), sudo=True)



import unittest
from nmigen import *
from nap import ZynqSocPlatform, SimPlatform
from nap.cores.axi import axil_read
from nap.cores.hdmi import generate_modeline, HdmiTx


class SocSmokeTest(unittest.TestCase):
    def test_hdmi_registers(self, testdata=0x1):
        platform = ZynqSocPlatform(SimPlatform())

        class Pins:
            def __init__(self):
                self.r = Signal()
                self.g = Signal()
                self.b = Signal()
                self.clock = Signal()

        dut = HdmiTx(Pins(), generate_clocks=False, modeline=generate_modeline(640, 480, 60))

        platform.add_sim_clock("pix", 117.5e6)

        def testbench():
            axi = platform.axi_lite_master
            memorymap = platform.memorymap
            for name, addr in memorymap.flattened.items():
                print(name, addr)
                yield from axil_read(axi, addr.address)
                # yield from axil_write(axi, addr.address, testdata)  # this will return an error if the register in question is a StatusRegister
                # self.assertEqual(testdata, (yield from axil_read(axi, addr.address)))

        platform.sim(dut, (testbench, "axi_lite"))



# TODO: implement atomic access (and before think about it)

from nmigen import *
from nmigen import Signal
from nap.soc import MemoryMap, Response, Peripheral
from nap.soc.csr_types import StatusSignal, ControlSignal, EventReg, _Csr

__all__ = ["CsrBank"]


class CsrBank(Elaboratable):
    def __init__(self):
        self.memorymap = MemoryMap()

    def reg(self, name: str, signal: _Csr):
        assert isinstance(signal, _Csr)
        writable = not isinstance(signal, StatusSignal)
        self.memorymap.allocate(name, writable, bits=len(signal), address=signal._address, obj=signal)

    def elaborate(self, platform):
        handled = Signal()

        def handle_read(m, addr, data, read_done):
            for iter_addr in range(0, self.memorymap.byte_len + 1, self.memorymap.bus_word_width_bytes):
                with m.If(iter_addr == addr):
                    for row in self.memorymap.direct_children:
                        bits_of_word = row.address.bits_of_word(iter_addr)
                        if bits_of_word:
                            word_range, signal_range = bits_of_word
                            if isinstance(row.obj, (ControlSignal, StatusSignal)):
                                m.d.sync += data[word_range.start:word_range.stop].eq(
                                    row.obj[signal_range.start:signal_range.stop]
                                )
                                read_done(Response.OK)
                                m.d.comb += handled.eq(1)
                            elif isinstance(row.obj, EventReg):
                                row.obj.handle_read(m, data[word_range.start:word_range.stop], read_done)
                                m.d.comb += handled.eq(1)
                            else:
                                raise NotImplementedError()
            with m.If(~handled):
                read_done(Response.ERR)

        def handle_write(m, addr, data, write_done):
            handled = Signal()
            for iter_addr in range(0, self.memorymap.byte_len + 1, self.memorymap.bus_word_width_bytes):
                with m.If(iter_addr == addr):
                    for row in self.memorymap.direct_children:
                        bits_of_word = row.address.bits_of_word(iter_addr)
                        if bits_of_word and row.writable:
                            word_range, signal_range = bits_of_word
                            if isinstance(row.obj, ControlSignal):
                                m.d.sync += row.obj[signal_range.start:signal_range.stop].eq(
                                    data[word_range.start:word_range.stop]
                                )
                                write_done(Response.OK)
                                m.d.comb += handled.eq(1)
                            elif isinstance(row.obj, StatusSignal):
                                write_done(Response.OK)
                                m.d.comb += handled.eq(1)
                            elif isinstance(row.obj, EventReg):
                                row.obj.handle_write(m, data[word_range.start:word_range.stop], write_done)
                                m.d.comb += handled.eq(1)
                            else:
                                raise NotImplementedError()
            with m.If(~handled):
                write_done(Response.ERR)

        m = Module()
        m.submodules += Peripheral(handle_read, handle_write, self.memorymap)
        return m



# this code is a translation of the original cvt.c from http://www.uruk.org/~erich/projects/cvt/cvt.c

# cvt.c  Generate mode timings using the CVT 1.1 Timing Standard
#  gcc cvt.c -O2 -o cvt -lm -Wall
#  November 2005:
# CVT revision 1.1 changes made by Erich Boleyn <erich@uruk.org>
# CVT is the VESA followon after GTF including new mode capabilities.
#  The CVT EXCEL(TM) SPREADSHEET, a sample (and the definitive)
# implementation of the CVT Timing Standard, is available at:
#  http://www.vesa.org/Public/CVT/CVTd6r1.xls
#  The GTF EXCEL(TM) SPREADSHEET can now be found at:
#  http://www.vesa.org/Public/GTF/GTF_V1R1.xls
#   Notes:
#   -- Didn't egregiously rewrite the program from scratch.
#   -- This is a different program than "gtf" because the CVT
#       timings are a bit different, and have some extra
#       features therein.
#   -- When producing interlaced modes, don't use non-intuitive
#       "full frame" frequency as the Vertical Frequency.  The more
#       intuitive definition is how often vertical retrace occurs.
#   Originally created from "gtf" written by Andy Ritger at NVidia.
# Original block comment at the beginning of the program follows:
# */
#
#/* gtf.c  Generate mode timings using the GTF Timing Standard
#  gcc gtf.c -o gtf -lm -Wall
#  Copyright (c) 2001, Andy Ritger  aritger@nvidia.com
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# o Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# o Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer
#   in the documentation and/or other materials provided with the
#   distribution.
# o Neither the name of NVIDIA nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
#  This program is based on the Generalized Timing Formula(GTF TM)
# Standard Version: 1.0, Revision: 1.0
#  The GTF Document contains the following Copyright information:
#  Copyright (c) 1994, 1995, 1996 - Video Electronics Standards
# Association. Duplication of this document within VESA member
# companies for review purposes is permitted. All other rights
# reserved.
#  While every precaution has been taken in the preparation
# of this standard, the Video Electronics Standards Association and
# its contributors assume no responsibility for errors or omissions,
# and make no warranties, expressed or implied, of functionality
# of suitability for any purpose. The sample code contained within
# this standard may be used without restriction.
#
#  The GTF EXCEL(TM) SPREADSHEET, a sample (and the definitive)
# implementation of the GTF Timing Standard, is available at:
#  ftp://ftp.vesa.org/pub/GTF/GTF_V1R1.xls
#    This program takes a desired resolution and vertical refresh rate,
# and computes mode timings according to the GTF Timing Standard.
# These mode timings can then be formatted as an XFree86 modeline
# or a mode description for use by fbset(8).
#    NOTES:
#  The GTF allows for computation of "margins" (the visible border
# surrounding the addressable video); on most non-overscan type
# systems, the margin period is zero.  I've implemented the margin
# computations but not enabled it because 1) I don't really have
# any experience with this, and 2) neither XFree86 modelines nor
# fbset fb.modes provide an obvious way for margin timings to be
# included in their mode descriptions (needs more investigation).
#
# The GTF provides for computation of interlaced mode timings;
# I've implemented the computations but not enabled them, yet.
# I should probably enable and test this at some point.
#
#  TODO:
#  o Add support for interlaced modes.
#  o Implement the other portions of the GTF: compute mode timings
#   given either the desired pixel clock or the desired horizontal
#   frequency.
#  o It would be nice if this were more general purpose to do things
#   outside the scope of the GTF: like generate double scan mode
#   timings, for example.
#
# o Printing digits to the right of the decimal point when the
#   digits are 0 annoys me.
#  o Error checking.

import math

CLOCK_STEP = 0.25  # Clock steps in MHz
MARGIN_PERCENT = 1.8  # % of active vertical image
H_SYNC_PER = 8.0  # sync % of horizontal image
CELL_GRAN = 8.4999  # assumed character cell granularity
CELL_GRAN_RND = 8.0  # assumed character cell granularity (round)
MIN_V_BPORCH = 3.0  # width of vsync in lines
MIN_V_PORCH_RND = 3.0  # width of vsync in lines
M = 600.0  # blanking formula gradient
C = 40.0  # blanking formula offset
K = 128.0  # blanking formula scaling factor
J = 20.0  # blanking formula scaling factor

# Standard Timing Parameters 
MIN_VSYNC_BP = 550.0  # min time of vsync + back porch (us)
H_SYNC_PERCENT = 8.0  # width of hsync as % of total line

# Reduced Blanking defines 
RB_MIN_V_BPORCH = 6.0  # lines
RB_V_FPORCH = 3.0  # lines
RB_MIN_V_BLANK = 460.0  # us
RB_H_SYNC = 32.0  # pixels
RB_H_BLANK = 160.0  # pixels

# C' and M' are part of the Blanking Duty Cycle computation 

C_PRIME = (((C - J) * K / 256.0) + J)
M_PRIME = (K / 256.0 * M)


#
#  - print the result of the named computation; this is
# useful when comparing against the CVT EXCEL spreadsheet.
#

#
# print_fb_mode() - print a mode description in fbset(8) format
# see the fb.modes(8) manpage.  The timing description used in
# this is rather odd; they use "left and right margin" to refer
# to the portion of the hblank before and after the sync pulse
# by conceptually wrapping the portion of the blank after the pulse
# to infront of the visible region; ie:
#
# Timing description I'm accustomed to:
# <--------1--------> <--2--> <--3--> <--4-->
# _________
# |-------------------|_______|       |_______
# R       SS      SE     FL
#
# 1: visible image
# 2: blank before sync (aka front porch)
# 3: sync pulse
# 4: blank after sync (aka back porch)
# R: Resolution
# SS: Sync Start
# SE: Sync End
# FL: Frame Length
# But the fb.modes format is:
# <--4--> <--------1--------> <--2--> <--3-->
# _________
# _______|-------------------|_______|       |
#
# The fb.modes(8) manpage refers to <4> and <2> as the left and
# right "margin" (as well as upper and lower margin in the vertical
# direction) -- note that this has nothing to do with the term
# "margin" used in the CVT Timing Standard.
# XXX always prints the 32 bit mode -- should I provide a command
# line option to specify the bpp?  It's simple enough for a user
# to edit the mode description after it's generated.
# 

#
# vert_refresh() - as defined by the CVT Timing Standard, compute the
# Stage 1 Parameters using the vertical refresh frequency.  In other
# words: input a desired resolution and desired refresh rate, and
# output the CVT mode timings.
# XXX margin computations are implemented but not tested (nor used by
# XFree86 of fbset mode descriptions, from what I can tell).


__all__ = ["generate_modeline"]


def generate_modeline(width, height, refresh, reduced_blanking=True):
    if (refresh % 60) != 0:
        reduced_blanking = False  # only possible for multiples of 60 Hz

    margins = 0
    interlaced = False

    #  1. Required Field Rate
    #       This is slightly different from the spreadsheet because we use
    #       a different result for interlaced video modes.  Simplifies this
    #       to the input field rate.
    #       [V FIELD RATE RQD] = [I/P FREQ RQD]
    #

    v_field_rate_rqd = refresh

    #  2. Horizontal Pixels
    #       In order to give correct results, the number of horizontal
    #       pixels requested is first processed to ensure that it is divisible
    #       by the character size, by rounding it to the nearest character
    #       cell boundary.
    #       [H PIXELS RND] = ((ROUNDDOWN([H PIXELS]/[CELL GRAN RND],0))
    #       *[CELLGRAN RND])
    #

    h_pixels_rnd = math.floor(width / CELL_GRAN_RND) * CELL_GRAN_RND

    #  2.5th Calculation, aspect_ratio & v_sync_rnd
    #       [ASPECT_RATIO] = IF(H_PIXELS_RND = CELL_GRAN_RND*ROUND((V_LINES*
    #       4.0/3.0)/CELL_GRAN_RND),"4:3")
    #       etc...
    #       [V_SYNC] = [value from table based on aspect ratio]
    #       [V_SYNC_RND] = ROUND(V_SYNC,0)  // Not needed in principle
    #

    if h_pixels_rnd == CELL_GRAN_RND * math.floor((height * 4.0 / 3.0) / CELL_GRAN_RND):
        aspect_ratio = "4:3"
        v_sync = 4
    elif h_pixels_rnd == CELL_GRAN_RND * math.floor((height * 16.0 / 9.0) / CELL_GRAN_RND):
        aspect_ratio = "16:9"
        v_sync = 5
    elif h_pixels_rnd == CELL_GRAN_RND * math.floor((height * 16.0 / 10.0) / CELL_GRAN_RND):
        aspect_ratio = "16:10"
        v_sync = 6
    elif h_pixels_rnd == CELL_GRAN_RND * math.floor((height * 5.0 / 4.0) / CELL_GRAN_RND):
        aspect_ratio = "5:4"
        v_sync = 7
    elif h_pixels_rnd == CELL_GRAN_RND * math.floor((height * 15.0 / 9.0) / CELL_GRAN_RND):
        aspect_ratio = "15:9"
        v_sync = 7
    else:
        # Default case of unknown aspect ratio
        aspect_ratio = "Custom"
        v_sync = 10

    v_sync_rnd = v_sync

    #
    # 3. Determine Left & Right Borders
    # Calculate the margins on the left and right side.
    # [LEFT MARGIN (PIXELS)] = (IF( [MARGINS RQD?]="Y",
    # (ROUNDDOWN( ([H PIXELS RND] * [MARGIN%] / 100 /
    # [CELL GRAN RND]),0)) * [CELL GRAN RND],
    # 0))
    # [RIGHT MARGIN (PIXELS)] = (IF( [MARGINS RQD?]="Y",
    # (ROUNDDOWN( ([H PIXELS RND] * [MARGIN%] / 100 /
    # [CELL GRAN RND]),0)) * [CELL GRAN RND],
    # 0))
    # 

    left_margin = math.floor(h_pixels_rnd * MARGIN_PERCENT / 100.0 / CELL_GRAN_RND) * CELL_GRAN_RND if margins else 0.0
    right_margin = left_margin

    #  4. Find total active pixels.
    #       Find total number of active pixels in image and left and right
    #       margins.
    #       [TOTAL ACTIVE PIXELS] = [H PIXELS RND] + [LEFT MARGIN (PIXELS)] +
    #       [RIGHT MARGIN (PIXELS)]
    #

    total_active_pixels = h_pixels_rnd + left_margin + right_margin

    #  5. Find number of lines per field.
    #       If interlace is requested, the number of vertical lines assumed
    #       by the calculation must be halved, as the computation calculates
    #       the number of vertical lines per field. In either case, the
    #       number of lines is rounded to the nearest integer.
    #
    #       [V LINES RND] = IF([INT RQD?]="y", ROUNDDOWN([V LINES]/2,0),
    #       ROUNDDOWN([V LINES],0))
    #

    v_lines_rnd = math.floor(height / 2.0) if interlaced else math.floor(height)

    #  6. Find Top and Bottom margins.
    #       [TOP MARGIN (LINES)] = IF([MARGINS RQD?]="Y",
    #       ROUNDDOWN(([MARGIN%]/100*[V LINES RND]),0),
    #       0)
    #       [BOT MARGIN (LINES)] = IF([MARGINS RQD?]="Y",
    #       ROUNDDOWN(([MARGIN%]/100*[V LINES RND]),0),
    #       0)
    #

    top_margin = math.floor(MARGIN_PERCENT / 100.0 * v_lines_rnd) if margins else (0.0)
    bot_margin = top_margin

    #  7. If interlace is required, then set variable [INTERLACE]=0.5:
    #       [INTERLACE]=(IF([INT RQD?]="y",0.5,0))
    #

    interlace = 0.5 if interlaced else 0.0

    # Here it diverges for "reduced blanking" or normal blanking modes.
    # 

    if reduced_blanking:
        h_blank = RB_H_BLANK

        #  8. Estimate Horiz. Period (us).
        #	   [H PERIOD EST] = ((1000000/V_FIELD_RATE_RQD)-RB_MIN_V_BLANK)/(V_LINES_RND+TOP_MARGIN+BOT_MARGIN)
        #

        h_period_est = (1000000.0 / v_field_rate_rqd - RB_MIN_V_BLANK) / (v_lines_rnd + top_margin + bot_margin)

        #  9. Find number of lines in vertical blanking.
        #	   [Actual VBI_LINES] = RB_MIN_V_BLANK/H_PERIOD_EST
        #	   [VBI_LINES] = ROUNDDOWN(RB_MIN_V_BLANK/H_PERIOD_EST,0) + 1
        #

        vbi_lines = RB_MIN_V_BLANK / h_period_est

        vbi_lines = math.floor(vbi_lines) + 1.0

        #  10. Check Vertical Blanking is sufficient.
        #	   [RB MIN VBI] = RB_V_FPORCH+V_SYNC_RND+RB_MIN_V_BPORCH
        #	   [ACT VBI LINES] = IF(VBI_LINES<RB_MIN_VBI,RB_MIN_VBI,VBI_LINES)
        #

        rb_min_vbi = RB_V_FPORCH + v_sync_rnd + RB_MIN_V_BPORCH
        act_vbi_lines = rb_min_vbi if (vbi_lines < rb_min_vbi) else vbi_lines

        #  11. Find total number of lines in vertical field.
        #	   [TOTAL V LINES] = ACT_VBI_LINES+V_LINES_RND+TOP_MARGIN+BOT_MARGIN+INTERLACE
        #

        total_v_lines = act_vbi_lines + v_lines_rnd + top_margin + bot_margin + interlace

        #  12. Find total number of pixels in a line (pixels).
        #	   [TOTAL PIXELS] = RB_H_BLANK+TOTAL_ACTIVE_PIXELS
        #

        total_pixels = total_active_pixels + RB_H_BLANK

        #  13. Find Pixel Clock Frequency (MHz).
        #	   [Non-rounded PIXEL_FREQ] = V_FIELD_RATE_RQD*TOTAL_V_LINES*TOTAL_PIXELS/1000000
        #	   [ACT PIXEL FREQ] = CLOCK_STEP * ROUND((V_FIELD_RATE_RQD*TOTAL_V_LINES*TOTAL_PIXELS/1000000)/CLOCK_STEP,0)
        #

        act_pixel_freq = v_field_rate_rqd * total_v_lines * total_pixels / 1000000.0

        act_pixel_freq = CLOCK_STEP * math.floor(act_pixel_freq / CLOCK_STEP)

        stage = 14

    else:  # Normal Blanking

        #  8. Estimate Horiz. Period (us).
        #	   [H PERIOD EST] = ((1/V_FIELD_RATE_RQD)-MIN_VSYNC_BP/1000000)/(V_LINES_RND+(2*TOP_MARGIN)+MIN_V_PORCH_RND+INTERLACE)*1000000
        #

        h_period_est = ((1 / v_field_rate_rqd) - MIN_VSYNC_BP / 1000000.0) / (
                v_lines_rnd + (2 * top_margin) + MIN_V_PORCH_RND + interlace) * 1000000.0

        #  9. Find number of lines in (SYNC + BACK PORCH).
        #	   [Estimated V_SYNC_BP] = ROUNDDOWN((MIN_VSYNC_BP/H_PERIOD_EST),0)+1
        #	   [Actual V_SYNC_BP] = MIN_VSYNC_BP/H_PERIOD_EST
        #	   [V_SYNC_BP] = IF(Estimated V_SYNC_BP<(V_SYNC+MIN_V_BPORCH),
        #	   V_SYNC+MIN_V_BPORCH,Estimated V_SYNC_BP)
        #

        v_sync_bp = MIN_VSYNC_BP / h_period_est

        v_sync_bp = math.floor(v_sync_bp) + 1

        v_sync_bp = v_sync + MIN_V_BPORCH if (v_sync_bp < v_sync + MIN_V_BPORCH) else v_sync_bp

        #  10. Find number of lines in back porch (Lines).
        #	   [Back porch] = V_SYNC_BP - V_SYNC_RND
        #

        #  11. Find total number of lines in vertical field.
        #	   [TOTAL V LINES] = V_LINES_RND+TOP_MARGIN+BOT_MARGIN
        #	   +V_SYNC_BP+INTERLACE+MIN_V_PORCH_RND
        #

        total_v_lines = v_lines_rnd + top_margin + bot_margin + v_sync_bp + interlace + MIN_V_PORCH_RND

        #  12. Find ideal blanking duty cycle from formula (%):
        #	   [IDEAL DUTY CYCLE] = C_PRIME-(M_PRIME*H_PERIOD_EST/1000)
        #

        ideal_duty_cycle = C_PRIME - (M_PRIME * h_period_est / 1000.0)

        #  13. Find blanking time to nearest cell (Pixels).
        #	   [H BLANK] = IF(IDEAL_DUTY_CYCLE<20,(ROUNDDOWN((TOTAL_ACTIVE_PIXELS*20/(100-20)/(2*CELL_GRAN_RND)),0))*(2*CELL_GRAN_RND),(ROUNDDOWN((TOTAL_ACTIVE_PIXELS*IDEAL_DUTY_CYCLE/(100-IDEAL_DUTY_CYCLE)/(2*CELL_GRAN_RND)),0))*(2*CELL_GRAN_RND))
        #

        cur_duty_cycle = 20.0 if (ideal_duty_cycle < 20.0) else ideal_duty_cycle
        h_blank = math.floor(
            (total_active_pixels * cur_duty_cycle / (100.0 - cur_duty_cycle) / (2.0 * CELL_GRAN_RND))) * (
                          2.0 * CELL_GRAN_RND)

        #  14. Find total number of pixels in a line (Pixels).
        #	   [TOTAL PIXELS] = TOTAL_ACTIVE_PIXELS + H_BLANK
        #

        total_pixels = total_active_pixels + h_blank

        #  15. Find pixel clock frequency (MHz).
        #	   [Non-rounded PIXEL FREQ] = TOTAL_PIXELS / H_PERIOD_EST
        #	   [ACT PIXEL FREQ] = CLOCK_STEP * ROUNDDOWN(
        #

        act_pixel_freq = total_pixels / h_period_est

        act_pixel_freq = CLOCK_STEP * math.floor(act_pixel_freq / CLOCK_STEP)

        stage = 16

    #  14/16. Find actual horizontal frequency (kHz)
    #       [ACT H FREQ] = 1000*ACT_PIXEL_FREQ/TOTAL_PIXELS
    #

    act_h_freq = 1000 * act_pixel_freq / total_pixels

    stage += 1

    #  15/17. Find actual field rate (Hz)
    #       [ACT FIELD RATE] = 1000*ACT_H_FREQ/TOTAL_V_LINES
    #

    act_field_rate = 1000 * act_h_freq / total_v_lines

    stage += 1

    #  16/18. Find actual vertical frame frequency (Hz)
    #       [ACT FRAME RATE] = IF(INT_RQD?=Y,ACT_FIELD_RATE/2,ACT_FIELD_RATE)
    #

    act_frame_rate = (act_field_rate / 2) if interlace else act_field_rate

    #
    # Extra computations not numbered in the CVT spreadsheet.
    # 

    #  20. Find Horizontal Back Porch.
    #       [H BACK PORCH] = H_BLANK/2
    #

    h_back_porch = h_blank / 2

    #  21. Find Horizontal Front Porch.
    #       [H SYNC RND] = IF(RED_BLANK_RQD?="Y",RB_H_SYNC,(ROUNDDOWN((H_SYNC_PER/100*TOTAL_PIXELS/CELL_GRAN_RND),0))*CELL_GRAN_RND)
    #

    if reduced_blanking:
        h_sync_rnd = RB_H_SYNC
    else:
        h_sync_rnd = math.floor(H_SYNC_PER / 100.0 * total_pixels / CELL_GRAN_RND) * CELL_GRAN_RND

    #  22. Find Horizontal Front Porch.
    #       [H FRONT PORCH] = H_BLANK - H_BACK_PORCH - H_SYNC_RND
    #

    h_front_porch = h_blank - h_back_porch - h_sync_rnd

    #  23. Find Vertical Front Porch.
    #       [V FRONT PORCH] = IF(RED_BLANK_RQD?="y",RB_V_FPORCH,MIN_V_PORCH_RND)
    #

    v_front_porch = RB_V_FPORCH if reduced_blanking else MIN_V_PORCH_RND

    return "Modeline \"{:d}x{:d}_{:.2f}{}{}\"  {:.2f}" "  {:d} {:d} {:d} {:d}" "  {:d} {:d} {:d} {:d}" "  {}{}HSync {}Vsync".format(
        int(h_pixels_rnd),
        int(height),
        refresh,
        "i" if interlaced else "",
        "R" if reduced_blanking else "",

        act_pixel_freq,

        int(h_pixels_rnd),
        int(h_pixels_rnd + h_front_porch),
        int(h_pixels_rnd + h_front_porch + h_sync_rnd),
        int(total_pixels),

        int(height),
        int(height + v_front_porch),
        int(height + v_front_porch + v_sync_rnd),
        int(total_v_lines - v_lines_rnd + height),

        "Interlace  " if interlaced else "",
        "+" if reduced_blanking else "-",
        "-" if reduced_blanking else "+",
    )



from nmigen.build import *
from nmigen.vendor.xilinx_7series import *

__all__ = ["ZyboPlatform"]


class ZyboPlatform(Xilinx7SeriesPlatform):
    device = "xc7z010"
    package = "clg400"
    speed = "1"
    resources = [
        Resource("hdmi", "north",
             # high speed serial lanes
             Subsignal("clock", DiffPairs("H16", "H17", dir='o'), Attrs(IOSTANDARD="TMDS_33")),
             Subsignal("b", DiffPairs("D19", "D20", dir='o'), Attrs(IOSTANDARD="TMDS_33")),
             Subsignal("g", DiffPairs("C20", "B20", dir='o'), Attrs(IOSTANDARD="TMDS_33")),
             Subsignal("r", DiffPairs("B19", "A20", dir='o'), Attrs(IOSTANDARD="TMDS_33")),
             Subsignal("out_en", Pins("F17", dir='o'), Attrs(IOSTANDARD="LVCMOS33")),
        )
    ]
    connectors = []



from abc import ABCMeta
from glob import glob
from importlib import import_module
from os.path import join, dirname, split

from nmigen import *
from nmigen.vendor.lattice_machxo_2_3l import LatticeMachXO2Platform
from nmigen.vendor.xilinx_7series import Xilinx7SeriesPlatform


class ImplementationMarkerMetaclass(ABCMeta):
    @property
    def implementation(self):
        if not hasattr(self, "_marker_type"):
            self._marker_type = type("{}Implementation".format(self.__name__), (), {})
        return self._marker_type


class PlatformAgnosticElaboratable(Elaboratable, metaclass=ImplementationMarkerMetaclass):
    """
    A helper to write Platform agnostic code. Searches in the vendor directories for the real elaboratable.
    """

    @classmethod
    def _search_in_path(cls, path):
        marker_type = cls.implementation
        basepath = __name__.split(".")[:-1] + [path]
        files = [split(p)[-1].replace(".py", "") for p in glob(join(dirname(__file__), path) + "/*.py")]
        for file in files:
            module = import_module(".".join(basepath + [file]))
            for candidate in [getattr(module, k) for k in dir(module)]:
                if isinstance(candidate, type) and issubclass(candidate, marker_type):
                    return candidate
        raise PrimitiveNotSupportedByPlatformError()

    def elaborate(self, platform):
        if isinstance(platform, Xilinx7SeriesPlatform):
            elaboratable = self._search_in_path("xilinx_s7")
        elif isinstance(platform, LatticeMachXO2Platform):
            elaboratable = self._search_in_path("lattice_machxo2")
        else:
            raise PlatformNotSupportedError()

        return elaboratable.elaborate(self, platform)


class PrimitiveNotSupportedByPlatformError(ValueError):
    pass


class PlatformNotSupportedError(ValueError):
    pass



if __name__ == "__main__":
    import os, atexit, code, readline, rlcompleter
    readline.parse_and_bind("tab: complete")

    class HistoryConsole(code.InteractiveConsole):
        def __init__(self, locals=None, filename="<console>",
                     histfile=os.path.expanduser("~/.pydriver-history")):
            code.InteractiveConsole.__init__(self, locals, filename)
            self.init_history(histfile)

        def init_history(self, histfile):
            readline.parse_and_bind("tab: complete")
            if hasattr(readline, "read_history_file"):
                try:
                    readline.read_history_file(histfile)
                except FileNotFoundError:
                    pass
                atexit.register(self.save_history, histfile)

        def save_history(self, histfile):
            readline.set_history_length(1000)
            readline.write_history_file(histfile)

    shell = HistoryConsole(locals())

    # setup the design variable
    design = Design(MemoryAccessor())

    print("welcome to the python shell to interact with the fpga")
    print("interact with it over the `design` variable and tab completion")
    shell.interact(banner="")



from .rgb import *
from .image_stream import *
from .image_convoluter import *
from .adapters import *
from .debayer import *
from .demo_source import *
from .focus_peeking import *
from .rearrange import *
from .resizer import *
from .test_util import *



import numpy as np
from matplotlib import pyplot as plt


def plt_discrete_hist(title, data):
    plt.figure()
    plt.title(title)
    min = np.min(data)
    max = np.max(data)
    plt.bar(np.arange(min, max + 1), np.bincount(np.ravel(data) - min, minlength=(max - min)), width=1.0)


def plt_hist(title, data, **kwargs):
    plt.figure()
    plt.title(title)
    plt.hist(np.ravel(data), **kwargs)


def plt_image(title, data, **kwargs):
    plt.figure()
    plt.title(title)
    plt.imshow(data, **kwargs)


def plt_show():
    plt.show()



import subprocess

CVT_BIN = "cvt"


def generate_modeline(width, height, refresh, reduced_blanking=True):
    if (refresh % 60) != 0:
        reduced_blanking = False  # only possible for multiples of 60 Hz

    cvt_bin = ([CVT_BIN] + ["-r"]) if reduced_blanking else [CVT_BIN]
    out, _ = subprocess.Popen(cvt_bin + [str(width), str(height), str(refresh)],
                              stdout=subprocess.PIPE).communicate()
    return out.split(b"\n")[1].decode("utf-8")



import copy
from itertools import chain, product
from math import log2

import numpy as np
from bitarray import bitarray
from numba import jit
from .py_wavelet_repack import pack, unpack
from .py_wavelet import ty
from huffman import codebook


def zero_rle(array, codebook):
    codebook = {**codebook, 1: 0}
    keys = np.array(sorted(codebook.keys(), reverse=True), dtype=ty)
    values = np.array([codebook[x] for x in keys], dtype=ty)
    return zero_rle_inner(array, keys, values)


@jit(nopython=True)
def zero_rle_inner(input_array, keys, values):
    output_array = np.zeros_like(input_array)

    zeroes = 0
    write_ptr = 0
    for v in input_array:
        if v == 0:
            zeroes += 1
        elif zeroes != 0:
            while zeroes != 0:
                for i, k in enumerate(keys):
                    if k <= zeroes:
                        output_array[write_ptr] = values[i]
                        write_ptr += 1
                        zeroes -= k
                        break
            output_array[write_ptr] = v
            write_ptr += 1
        else:
            output_array[write_ptr] = v
            write_ptr += 1
    while zeroes != 0:
        for i, k in enumerate(keys):
            if k <= zeroes:
                output_array[write_ptr] = values[i]
                write_ptr += 1
                zeroes -= k

    return output_array[:write_ptr]


@jit(nopython=True)
def n_combine(input_array, n, first_symbol):
    output_array = np.zeros_like(input_array)
    write_ptr = 0
    in_range_cnt = 0
    symbol = 0
    for i in range(len(input_array)):
        elem = input_array[i]
        if -1 <= elem <= +1:
            symbol += (in_range_cnt * 3) * (elem + 1)
            if in_range_cnt == n - 1:
                output_array[write_ptr] = first_symbol + symbol
                write_ptr += 1
                in_range_cnt = 0
                symbol = 0
            else:
                in_range_cnt += 1
        else:
            for x in range(in_range_cnt + 1):
                output_array[write_ptr] = input_array[i - x]
                write_ptr += 1
            in_range_cnt = 0
            symbol = 0
    return output_array[:write_ptr]


def zero_rle_decode(array, codebook, length):
    result = np.empty(length, dtype=ty)
    # This assumes that the rle symbols are continuous starting with the one with the lowest value
    codebook_start = min(codebook.values())
    codebook_list = np.array(list(codebook.keys()), dtype=ty)
    read = zero_rle_decode_inner(array, result, codebook_list, codebook_start)
    if read <= 0:
        assert False
    return result, read


@jit('int32(int32[:], int32[:], int32[:], int32)', nopython=True)
def zero_rle_decode_inner(input_array, output_array, codebook, codebook_start):
    target_write_index = len(output_array)
    write_index = 0
    for i, v in enumerate(input_array):
        # This assumes that the rle symbols are above the range of allowed normal codewords
        if v < codebook_start:
            output_array[write_index] = v
            write_index += 1
        else:
            n_zeros = codebook[v - codebook_start]
            for _ in range(n_zeros):
                output_array[write_index] = 0
                write_index += 1
        if write_index > target_write_index:
            return -42
        if write_index == target_write_index:
            return i + 1
    return -write_index


def possible_region_codes(levels=3):
    return list(chain(*[
        [1],
        *[[l * 10 + 2, l * 10 + 3, l * 10 + 4] for l in range(1, levels + 1)],
    ]))


def fill_reference_frame(h, w, levels=3):
    ref = np.zeros((h, w), dtype=ty)
    ref[:h // 2, :w // 2] = 1 if levels == 1 else fill_reference_frame(h // 2, w // 2, levels - 1)
    ref[:h // 2, w // 2:] = levels * 10 + 2
    ref[h // 2:, :w // 2] = levels * 10 + 3
    ref[h // 2:, w // 2:] = levels * 10 + 4
    return ref


class NumericRange:
    def __init__(self, min, max):
        self.min = int(np.sign(min) * np.ceil(np.abs(min)))
        self.max = int(np.sign(max) * np.ceil(np.abs(max)))

    def __repr__(self):
        return f'NumericRange({self.min}, {self.max})'

    def _compute(self, operation, other):
        if isinstance(other, NumericRange):
            other_list = [other.min, other.max]
        else:
            other_list = [other]
        values = [getattr(float(a), operation)(float(b)) for a, b in product([self.min, self.max], other_list)]
        return NumericRange(min(values), max(values))

    def __neg__(self):
        return NumericRange(-self.max, -self.min)

    def __add__(self, other):
        return self._compute('__add__', other)

    def __sub__(self, other):
        return self._compute('__sub__', other)

    def __mul__(self, other):
        return self._compute("__mul__", other)

    def __truediv__(self, other):
        return self._compute("__truediv__", other)

    def __floordiv__(self, other):
        return self._compute("__floordiv__", other)


def numeric_range_from_region_code(region_code, levels, input_range, quantization):
    lf = lambda nr: nr + nr
    hf = lambda nr: (nr - nr) + (-nr - nr + nr + nr + 4) // 8

    lf_h = lambda level: lf(input_range) if level == 1 else lf(lf_v_quantized(level - 1))
    lf_v = lambda level: lf(lf_h(level))
    lf_v_quantized = lambda level: lf_v(level) / quantization[levels - level - 1][0]
    hf_top_right = lambda level: hf(lf_h(level))
    hf_bottom_left = lambda level: hf(lf_v(level))
    hf_bottom_right = lambda level: hf(hf_top_right(level))

    if region_code == 1:
        return lf_v_quantized(levels)

    level = (levels - region_code // 10) + 1
    if region_code % 10 == 2:
        return hf_top_right(level) / quantization[level - 1][1]
    elif region_code % 10 == 3:
        return hf_bottom_left(level) / quantization[level - 1][2]
    elif region_code % 10 == 4:
        return hf_bottom_right(level) / quantization[level - 1][3]


def numeric_range_from_region_code_with_rle(region_code, levels, input_range, quantization):
    nr = numeric_range_from_region_code(region_code, levels, input_range, quantization)
    return NumericRange(nr.min, nr.max + len(gen_rle_dict(region_code, levels, input_range, quantization)) + (3**2) + (3 ** 3))


def gen_rle_dict(region_code, levels, input_range, quantization):
    rle_codes = [4, 5, 6, 7, 8, 10, 12, 15, 18, 25, 35, 50]
    nr = numeric_range_from_region_code(region_code, levels, input_range, quantization)
    return {v: i + nr.max for i, v in enumerate(rle_codes)}


def rle_region(data, region_code, levels, input_range, quantization):
    nr = numeric_range_from_region_code(region_code, levels, input_range, quantization)
    outliers = data[np.where((data < nr.min) | (data > nr.max))]
    if len(outliers) > 0:
        raise ValueError
    if region_code == 1:
        return data  # don't do rle for lf data
    else:
        rled_region = zero_rle(data, gen_rle_dict(region_code, levels, input_range, quantization))
        highest_rle_symbol = nr.max + len(gen_rle_dict(region_code, levels, input_range, quantization))
        merge_3 = n_combine(rled_region, 3, highest_rle_symbol)
        merge_2 = n_combine(merge_3, 2, highest_rle_symbol + (3 ** 3))
        return merge_2


def rle_region_decode(data, region_code, levels, length, input_range, quantization):
    numeric_range = numeric_range_from_region_code_with_rle(region_code, levels, input_range, quantization)
    outliers, = np.where((data < numeric_range.min) | (data > numeric_range.max))
    read_length = length
    if len(outliers) > 0:
        read_length = np.min(outliers)
    assert read_length > 0
    if region_code == 1:
        return data, length
    else:
        return zero_rle_decode(data[:read_length], gen_rle_dict(region_code, levels, input_range, quantization), length)


def to_chunks(image, levels):
    reference = fill_reference_frame(*image.shape, levels)
    packed_reference = pack(reference, levels)
    packed_image = pack(image, levels)

    for ref_line, real_line in list(zip(packed_reference, packed_image)):
        regions = np.unique(ref_line)
        for region_code in regions:
            if region_code == 0:
                continue
            yield region_code, real_line[np.where(ref_line == region_code)]


def rle_compress_chunks(chunks, levels, input_range, quantization):
    for rc, data in chunks:
        yield rle_region(data, rc, levels, input_range, quantization)


def empty_symbol_frequencies_dict(levels, input_range, quantization):
    regions = possible_region_codes(levels)
    numeric_range = [numeric_range_from_region_code_with_rle(rc, levels, input_range, quantization) for rc in regions]
    return {rc: np.zeros(nr.max - nr.min + 1) for nr, rc in zip(numeric_range, regions)}


def compute_symbol_frequencies(region_codes, compressed_chunks, levels, input_range, quantization):
    symbol_frequencies = empty_symbol_frequencies_dict(levels, input_range, quantization)
    for compressed_chunk, rc in zip(compressed_chunks, region_codes):
        nr = numeric_range_from_region_code_with_rle(rc, levels, input_range, quantization)
        symbol_frequencies[rc] += np.bincount(compressed_chunk - nr.min, minlength=(nr.max - nr.min + 1))

    assert np.sum(np.concatenate(list(symbol_frequencies.values()))) == np.concatenate(compressed_chunks).size
    return symbol_frequencies


def merge_symbol_frequencies(symbol_frequencies_list):
    head, *rest = symbol_frequencies_list
    result = copy.deepcopy(head)
    for other in rest:
        for k in result.keys():
            result[k] += other[k]
    return result


def generate_huffman_tables(symbol_frequencies, levels, input_range, quantization, max_table_size=1024):
    to_return = {}
    for rc, frequencies in symbol_frequencies.items():
        nr = numeric_range_from_region_code_with_rle(rc, levels, input_range, quantization)
        symbols = np.arange(nr.min, nr.max + 1, dtype=ty)
        if rc == 1:
            cb = {}
        else:
            sorting_indecies = np.argsort(frequencies)

            escape_symbol = nr.max + 1
            real_symbols = np.append(symbols[sorting_indecies[-max_table_size:]], [escape_symbol])

            escape_frequency = np.sum(frequencies[:-max_table_size])
            real_frequencies = np.append(frequencies[sorting_indecies[-max_table_size:]], [escape_frequency])

            cb = codebook(zip(real_symbols, real_frequencies))

        to_return[rc] = {k: bitarray(v) for k, v in cb.items()}

    return to_return


def get_huffman_size(huffman_tables, region_codes, rle_chunks, levels, input_range, quantization):
    huffman_length_arrays = {}
    for rc, huffman_table in huffman_tables.items():
        nr = numeric_range_from_region_code_with_rle(rc, levels, input_range, quantization)
        bits_needed = int(np.ceil(log2(nr.max - nr.min + 1)))
        if rc == 1:
            lengths = np.full(nr.max - nr.min + 1, bits_needed, dtype=np.uint64)
        else:
            escape_symbol = nr.max + 1
            escape_symbol_length = len(huffman_table[escape_symbol])
            lengths = np.full(nr.max - nr.min + 1, bits_needed + escape_symbol_length, dtype=np.uint64)
            for symbol, code in huffman_table.items():
                if symbol != escape_symbol:
                    lengths[symbol - nr.min] = len(code)
        huffman_length_arrays[rc] = lengths

    size = 0
    for rc, data in zip(region_codes, rle_chunks):
        nr = numeric_range_from_region_code_with_rle(rc, levels, input_range, quantization)
        size += np.sum(huffman_length_arrays[rc][data - nr.min])
    return size


def huffman_encode(huffman_tables, region_codes, rle_chunks):
    huffman_encoded = bitarray()
    for rc, data in zip(region_codes, rle_chunks):
        huffman_encoded.encode(huffman_tables[rc], data)
    return huffman_encoded


def compress(image, levels, input_range, quantization):
    chunks = list(to_chunks(image, levels))
    region_codes, uncompressed_chunks = zip(*chunks)
    rle_chunks = list(rle_compress_chunks(chunks, levels, input_range, quantization))

    non_compressed = np.concatenate(list(uncompressed_chunks))
    rle_compressed = np.concatenate(list(rle_chunks))
    rle_ratio = len(non_compressed) / len(rle_compressed)

    symbol_frequencies = compute_symbol_frequencies(region_codes, rle_chunks, levels, input_range, quantization)
    huffman_tables = generate_huffman_tables(symbol_frequencies, levels, input_range, quantization)
    huffman_encoded = huffman_encode(huffman_tables, region_codes, rle_chunks)

    huffman_ratio = len(rle_compressed) / len(huffman_encoded.tobytes())
    total_ratio = image.size * (log2(input_range.max + 1) / 8) / len(huffman_encoded.tobytes())

    return rle_compressed, symbol_frequencies, rle_ratio, huffman_ratio, total_ratio


def uncompress(original_shape, compressed, levels, bit_depth, quantization):
    reference = fill_reference_frame(*original_shape, levels)
    packed_reference = pack(reference, levels)
    rle_decompressed_frame = pack(fill_reference_frame(*original_shape, levels), levels)
    rle_compressed_ptr = 0
    for ref_line, line in list(zip(packed_reference, rle_decompressed_frame)):
        regions = np.unique(ref_line)
        for region_code in regions:
            if region_code == 0:
                continue
            region_range, = np.where(ref_line == region_code)
            start, end = np.min(region_range), np.max(region_range) + 1
            region_len = end - start

            rle_slice = compressed[rle_compressed_ptr:rle_compressed_ptr + region_len]
            rle_decoded, consumed = rle_region_decode(rle_slice, region_code, levels, bit_depth, region_len, quantization)
            line[start:end] = rle_decoded
            rle_compressed_ptr += consumed

    return unpack(rle_decompressed_frame, levels)



import unittest

from nmigen import *
from nmigen.sim import Simulator

from .nmigen_misc import nMin, nAny, nAll, nMax, assert_is_pot, log2, ends_with, nAbsDifference


class TestNMigenMisc(unittest.TestCase):
    def _resolve(self, expr):
        sim = Simulator(Module())

        a = []

        def testbench():
            a.append((yield expr))

        sim.add_process(testbench)
        sim.run()
        return a[0]

    def test_nMin(self):
        self.assertEqual(3, self._resolve(nMin(3, 7)))
        self.assertEqual(3, self._resolve(nMin(7, 3)))

    def test_nMax(self):
        self.assertEqual(7, self._resolve(nMax(3, 7)))
        self.assertEqual(7, self._resolve(nMax(7, 3)))

    def test_nAny(self):
        self.assertEqual(1, self._resolve(nAny([Const(1), Const(0), Const(0), Const(0)])))
        self.assertEqual(1, self._resolve(nAny([Const(0), Const(0), Const(1), Const(1)])))
        self.assertEqual(0, self._resolve(nAny([Const(0), Const(0), Const(0), Const(0)])))

    def test_nAll(self):
        self.assertEqual(1, self._resolve(nAll([Const(1), Const(1), Const(1), Const(1)])))
        self.assertEqual(0, self._resolve(nAll([Const(0), Const(0), Const(1), Const(1)])))
        self.assertEqual(0, self._resolve(nAll([Const(0), Const(0), Const(0), Const(0)])))

    def test_is_pot(self):
        assert_is_pot(2)
        assert_is_pot(4)
        assert_is_pot(64)
        assert_is_pot(512)
        with self.assertRaisesRegex(AssertionError, "is not a power of two"):
            assert_is_pot(7)
        with self.assertRaisesRegex(AssertionError, "is not a power of two"):
            assert_is_pot(42)
        with self.assertRaisesRegex(AssertionError, "is not a power of two"):
            assert_is_pot(196)

    def test_log2(self):
        self.assertEqual(1, log2(2))
        self.assertEqual(2, log2(4))
        self.assertEqual(10, log2(1024))
        with self.assertRaisesRegex(AssertionError, "is not a power of two"):
            assert_is_pot(196)

    def test_ends_with(self):
        self.assertTrue(self._resolve(ends_with(Const(0b0001, 4), "01")))
        self.assertFalse(self._resolve(ends_with(Const(0b0001, 4), "10")))
        self.assertTrue(self._resolve(ends_with(Const(0b10001011, 4), "1011")))
        self.assertTrue(self._resolve(ends_with(Const(0b10001011, 4), "011")))
        self.assertFalse(self._resolve(ends_with(Const(0b10001011, 4), "0110")))

    def test_nAbsDifference(self):
        self.assertEqual(7, self._resolve(nAbsDifference(10, 3)))
        self.assertEqual(7, self._resolve(nAbsDifference(3, 10)))
        self.assertEqual(3, self._resolve(nAbsDifference(12, 9)))
        self.assertEqual(3, self._resolve(nAbsDifference(9, 12)))



from nmigen import *
import unittest
from nmigen.sim import Passive
from nap import ImageStream, write_to_stream, write_frame_to_stream, read_frame_from_stream, crop, SimPlatform
from .image_convoluter import ImageProxy, ImageConvoluter


class ImageProxyTest(unittest.TestCase):
    def test_request(self):
        x = Signal(16)
        y = Signal(16)
        image = ImageProxy(12, x, y)
        px = image[x + 1, y + 1]
        self.assertIsInstance(px, Signal)
        self.assertEqual(image.offset_mapping, {(-1, -1): px})

    def test_request_zero(self):
        x = Signal(16)
        y = Signal(16)
        image = ImageProxy(12, x, y)
        px = image[x, y]
        self.assertIsInstance(px, Signal)
        self.assertEqual(image.offset_mapping, {(0, 0): px})

    def test_request_neg(self):
        x = Signal(16)
        y = Signal(16)
        image = ImageProxy(12, x, y)
        px = image[x - 1, y - 1]
        self.assertIsInstance(px, Signal)
        self.assertEqual(image.offset_mapping, {(1, 1): px})

    def test_request_equal(self):
        x = Signal(16)
        y = Signal(16)
        image = ImageProxy(12, x, y)
        self.assertIs(image[x + 1, y + 1], image[x + 1, y + 1])
        self.assertIsNot(image[x + 1, y + 2], image[x + 1, y + 1])

    def test_illegal_expr(self):
        x = Signal(16)
        y = Signal(16)
        image = ImageProxy(12, x, y)

        with self.assertRaises(ValueError):
            px = image[0, 0]
        with self.assertRaises(ValueError):
            px = image[x * 2, y * 2]
        with self.assertRaises(ValueError):
            px = image[y, x]


class VideoTransformerTest(unittest.TestCase):
    def check_move_transformer(self, transform_xy, testdata, testdata_transformed, crop_top=0, crop_left=0, crop_bottom=0, crop_right=0):
        m = Module()
        tx, ty = transform_xy

        def transformer_function(x, y, image):
            return image[x + tx, y + ty]

        input = ImageStream(32)
        transformer = m.submodules.transformer = ImageConvoluter(input, transformer_function, 10, 10)

        def write_process():
            yield from write_frame_to_stream(input, testdata, pause=True)
            yield Passive()
            while True:
                yield from write_to_stream(input, line_last=0, frame_last=0, payload=0)

        def read_process():
            self.assertEqual(crop((yield from read_frame_from_stream(transformer.output, pause=True)), left=crop_left, right=crop_right, bottom=crop_bottom, top=crop_top),
                             testdata_transformed)

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(write_process, "sync")
        platform.sim(m, read_process)

    def test_passthrough_transformer(self):
        testdata = [[x * y for x in range(10)] for y in range(10)]
        self.check_move_transformer((0, 0), testdata, testdata)

    def test_shift_1x_negative_transformer(self):
        testdata = [[x * y for x in range(10)] for y in range(10)]
        self.check_move_transformer(
            (-1, 0),
            testdata,
            [[px for px in line[:-1]] for line in testdata],
            crop_left=1
        )

    def test_shift_1y_negative_transformer(self):
        testdata = [[x * y for x in range(10)] for y in range(10)]
        self.check_move_transformer(
            (0, -1),
            testdata,
            [[0] * 10] + [[px for px in line] for line in testdata[:-1]]
        )

    def test_shift_1x_positive_transformer(self):
        testdata = [[x * y for x in range(10)] for y in range(10)]
        self.check_move_transformer(
            (+1, 0),
            testdata,
            [[px for px in line[1:]] + [0] for line in testdata]
        )

    def test_shift_1y_positive_transformer(self):
        testdata = [[x * y for x in range(10)] for y in range(10)]
        self.check_move_transformer(
            (0, +1),
            testdata,
            [[px for px in line] for line in testdata[1:] + [[0] * 10]]
        )

    def check_non_moving_xy(self, transformer_function, crop_top=0, crop_left=0, crop_bottom=0, crop_right=0):
        m = Module()

        width, height = 9, 9
        input = ImageStream(32)
        transformer = m.submodules.transformer = ImageConvoluter(input, transformer_function, width, height)

        def write_process():
            testdata = [[x * y for x in range(width)] for y in range(height)]
            yield from write_frame_to_stream(input, testdata, pause=True)
            yield from write_frame_to_stream(input, testdata, pause=True)
            yield from write_frame_to_stream(input, testdata, pause=True)
            yield Passive()
            while True:
                yield from write_to_stream(input, line_last=0, frame_last=0, payload=0)

        def read_process():
            (yield from read_frame_from_stream(transformer.output, pause=True))
            first = crop((yield from read_frame_from_stream(transformer.output, pause=True)), left=crop_left, right=crop_right, bottom=crop_bottom, top=crop_top)
            second = crop((yield from read_frame_from_stream(transformer.output, pause=True)), left=crop_left, right=crop_right, bottom=crop_bottom, top=crop_top)
            self.assertEqual(first, second)

        platform = SimPlatform()
        platform.add_sim_clock("sync", 100e6)
        platform.add_process(write_process, "sync")
        platform.sim(m, read_process)

    def test_non_moving_xy_pattern(self):
        def transformer_function(x, y, image):
            return (x % 2 == 0) & (y % 2 == 0)
        self.check_non_moving_xy(transformer_function)

    def test_non_moving_xy_passthrough(self):
        def transformer_function(x, y, image):
            return image[x, y]
        self.check_non_moving_xy(transformer_function)

    def test_non_moving_xy_shift_positive(self):
        def transformer_function(x, y, image):
            return image[x+1, y+2]
        self.check_non_moving_xy(transformer_function, crop_right=1, crop_bottom=2)

    def test_non_moving_xy_shift_negative(self):
        def transformer_function(x, y, image):
            return image[x-2, y-1]
        self.check_non_moving_xy(transformer_function)